import os
import logging
import urllib
import time
import requests
from datetime import datetime
from itertools import islice
from abc import ABC, abstractmethod

UNINSTALLED_CONNECTORS = []

# common dependencies to snowflake and bigquery connectors
try:
    from google.cloud import storage
    from google.oauth2 import service_account
except ImportError:
    UNINSTALLED_CONNECTORS += ['bigquery', 'snowflake']

try:
    from google.cloud import bigquery
except ImportError:
    UNINSTALLED_CONNECTORS += ['bigquery']

try:
    import googleapiclient.discovery
    from google.api_core.exceptions import PreconditionFailed
    import snowflake.connector
except ImportError:
    UNINSTALLED_CONNECTORS += ['snowflake']

UNINSTALLED_CONNECTORS = list(set(UNINSTALLED_CONNECTORS))

logger = logging.getLogger('tinybird-connect')


def _now():
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


def _log(text):
    logger.info(f"{_now()} - {text}")


class GCS():
    MAX_COMPOSE = 31
    READ_TIMEOUT = 3

    def __init__(self, options):
        self.options = options
        filename = options['service_account']
        if not filename:
            filename = os.getenv('GOOGLE_APPLICATION_CREDENTIALS')
        self.credentials = service_account.Credentials.from_service_account_file(
            filename=filename,
            scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )
        self.storage_client = storage.Client(project=options['project_id'], credentials=self.credentials)

    def gcs_url(self):
        return f'gcs://{self.bucket_name()}/'

    def gs_url(self):
        return f'gs://{self.bucket_name()}/'

    def bucket_name(self):
        return self.options['bucket_name']

    def sign(self, blob_name):
        bucket = self.storage_client.get_bucket(self.bucket_name())
        blob = bucket.get_blob(blob_name)
        if not blob:
            raise Exception('Warning: File not found. This probably just means there is no new data to load')
        return blob.generate_signed_url(expiration=86400, version='v4', scheme='https')

    def grant_access(self, member_name):
        role_name = self.get_role_name()
        service = googleapiclient.discovery.build(
            "iam", "v1", credentials=self.credentials
        )
        result = service.roles().list(parent='projects/' + self.options['project_id']).execute()
        exists = False if 'roles' not in result else any(role['title'] == role_name for role in result['roles'])
        if not exists:
            role_name = self.create_custom_role(service)
        self.add_member_to_bucket(member_name, role_name, self.bucket_name())

    def get_role_name(self):
        return f"{self.bucket_name().replace('-', '')}_5".lower()

    def create_custom_role(self, service):
        service.projects().roles().create(
            parent=f"projects/{self.options['project_id']}",
            body={
                'roleId': self.get_role_name(),
                'role': {
                    'title': self.get_role_name(),
                    'description': self.get_role_name(),
                    'includedPermissions': ['storage.objects.create', 'storage.buckets.get', 'storage.objects.delete', 'storage.objects.get', 'storage.objects.list'],
                    'stage': 'GA'
                }
            }).execute()
        return self.get_role_name()

    def add_member_to_bucket(self, member_name, iam_role, bucket_name, retry=True):
        try:
            project_id = self.options['project_id']
            role = f'projects/{project_id}/roles/{iam_role}'
            member = f'serviceAccount:{member_name}'

            bucket = self.storage_client.bucket(bucket_name)
            policy = bucket.get_iam_policy(requested_policy_version=3)
            policy.bindings.append({"role": role, "members": {member}})
            bucket.set_iam_policy(policy)
        except PreconditionFailed as e:
            _log(str(e))
            if retry:
                _log('retrying...')
                self.add_member_to_bucket(member_name, iam_role, bucket_name, retry=False)

    def compose(self, source_prefix, destination_blob_name):
        bucket = self.storage_client.bucket(self.bucket_name())
        retries = 0
        sources = None
        # for some reason when this list_blobs run in GH actions takes forever and fails
        # so retry many times with a lower timeout
        while not sources and retries < 5:
            try:
                sources = list(self.storage_client.list_blobs(
                    self.bucket_name(), prefix=source_prefix, timeout=self.READ_TIMEOUT, fields='items(name)'
                ))
            except requests.exceptions.ReadTimeout:
                _log("exception fetching blob list from GCS")
            retries += 1

        if sources is None:
            raise Exception("couldn't load list of blobs from GCS")

        _log(f'Exported {len(sources)} CSV files')
        if len(sources) == 0:
            raise Exception('Warning: Nothing to load. Aborting next steps')

        sources_iter = iter(sources)
        chunks = [list(islice(sources_iter, self.MAX_COMPOSE)) for i in range(0, 1 + int(len(sources) / self.MAX_COMPOSE))]
        blobs_to_delete = []
        if len(chunks) == 1:
            _log(f"Composing {destination_blob_name} from {','.join([c.name for c in chunks[0]])}")
            destination = bucket.blob(destination_blob_name)
            destination.content_type = "text/csv"
            destination.compose(chunks[0])
            blobs_to_delete += chunks[0]
        else:
            count = 0
            for chunk in chunks:
                destination_name = f'{destination_blob_name}_part{count}'
                if count > 0:
                    blob = bucket.get_blob(f'{destination_blob_name}_part{count - 1}')
                    chunk.append(blob)
                if count == len(chunks) - 1:
                    destination_name = destination_blob_name
                _log(f"[{count}] - Composing {destination_name} from {','.join([c.name for c in chunk])}")
                destination = bucket.blob(destination_name)
                destination.content_type = "text/csv"
                destination.compose(chunk)
                count = count + 1
                blobs_to_delete += chunk

        try:
            _log('Removing temp blobs')
            for blob in blobs_to_delete:
                blob.delete()
        except Exception:
            pass

        _log('Done composing!')

    def delete(self, blob_name):
        bucket = self.storage_client.bucket(self.bucket_name())
        bucket.delete_blob(blob_name)


class Connector(ABC):

    def __init__(self, options):
        self.options = options
        self.options['gcs'] = {
            'project_id': options['project_id'],
            'service_account': options['service_account'],
            'sign_service_account': options['service_account'],
            'bucket_name': options['bucket_name'],
        }
        self.gcs = GCS(self.options['gcs'])

    @abstractmethod
    def export_to_gcs(self, sql, destination):
        pass

    @abstractmethod
    def clean(self, blob_name):
        pass


class BigQuery(Connector):
    def __init__(self, options):
        super().__init__(options)
        self.configure()

    def configure(self):
        pass

    def datasource_analyze(self, resource):
        raise Exception("BigQuery does not support data source analyze")

    def export_to_gcs(self, sql, destination):
        mm = str(int(round(time.time() * 1000)))
        destination = f"{destination}_{mm}"
        with_headers: bool = self.options.get('with_headers')
        sql = f"""
                EXPORT DATA OPTIONS(
                    uri='{self.gcs.gs_url()}{destination}*.csv',
                    format='CSV',
                    overwrite=true,
                    header={ "true" if with_headers else "false" },
                    field_delimiter=',') AS
                    {sql}
                """
        self.execute(sql)
        final_destination = f'{destination}_final.csv'
        self.gcs.compose(f'{destination}', final_destination)
        return urllib.parse.unquote(self.gcs.sign(final_destination))

    def execute(self, sql, result=False):
        client = self.connector()
        job_config = bigquery.QueryJobConfig()

        query_job = client.query(
            sql,
            job_config=job_config,
        )

        return query_job.result()  # Waits for the query to finish

    def connector(self):
        credentials = service_account.Credentials.from_service_account_file(
            filename=self.options['service_account'],
            scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )
        self.credentials = credentials

        return bigquery.Client(credentials=credentials, project=credentials.project_id)

    def clean(self, blob_name):
        self.gcs.delete(blob_name)


class Snowflake(Connector):
    DEFAULT_STORAGE_INTEGRATION = 'tb__gcs_int'
    DEFAULT_STAGE = 'tb__gcs_stage'

    def __init__(self, options):
        super().__init__(options)
        self.configure()

    def configure(self):
        self.create_storage_integration()
        self.create_stage()
        self.gcs.grant_access(self.get_sf_member_name())

    def create_storage_integration(self):
        sql = f"""
                    create storage integration {self.storage_integration()}
                        type = external_stage
                        storage_provider = gcs
                        enabled = true
                        storage_allowed_locations = ('{self.gcs.gcs_url()}');
                """
        try:
            self.execute(sql)
        except Exception as e:
            if 'already exists' in str(e):
                pass
            else:
                raise e

    def get_sf_member_name(self):
        try:
            sql = F'DESC STORAGE INTEGRATION {self.storage_integration()};'
            result = self.execute(sql, result=True)
            for row in result:
                if row['property'] == 'STORAGE_GCP_SERVICE_ACCOUNT':
                    return row['property_value']
        except Exception:
            return None

    def create_stage(self):
        sql = f"""
                    create stage "{self.options['schema']}".{self.stage()}
                        url='{self.gcs.gcs_url()}'
                        storage_integration = {self.storage_integration()};
                """
        try:
            self.execute(sql)
        except Exception as e:
            if 'already exists' in str(e):
                pass

    def export_to_gcs(self, sql, destination):
        mm = str(int(round(time.time() * 1000)))
        destination = f"{destination}_{mm}"
        with_headers: bool = self.options.get('with_headers')
        sql = f"""copy into '@{self.stage()}/{destination}'
                    from ({sql})
                    overwrite = true
                    file_format = (TYPE=CSV COMPRESSION=NONE ESCAPE_UNENCLOSED_FIELD=NONE FIELD_DELIMITER='|' FIELD_OPTIONALLY_ENCLOSED_BY='"' null_if=())
                    header = {"true" if with_headers else "false" }
                    max_file_size = 5000000000;
                """
        self.execute(sql)
        final_destination = f'{destination}_final.csv'
        self.gcs.compose(destination, final_destination)
        return urllib.parse.unquote(self.gcs.sign(final_destination))

    def execute(self, sql, result=False):
        ctx = self.connector()
        cs = ctx.cursor(snowflake.connector.DictCursor)
        try:
            cs.execute(f"use role {self.options['role']};")
            cs.execute(f"use warehouse {self.options['warehouse']};")
            cs.execute(sql)
            if result:
                result = cs.fetchall()
        finally:
            cs.close()
        ctx.close()
        return result

    def datasource_analyze(self, resource):
        """returns .datasource file for resource"""
        # TODO: deal with the right precisions for Integers
        YES = 'Y'
        NO = 'N'

        def from_snowflake_type(t, nullable=NO):
            """transforms snowflake types to CH ones"""
            the_type = 'String'
            if t.startswith('NUMBER'):
                the_type = 'Int32'
            if t.startswith('FLOAT') or t.startswith('DOUBLE') or t.startswith('REAL') or t.startswith('NUMERIC') or t.startswith('DECIMAL'):
                the_type = 'Float32'
            if t == 'DATE':
                the_type = 'Date'
            if t == 'DATETIME' or t.startswith('TIMESTAMP'):
                the_type = 'DateTime'
            if nullable == YES:
                the_type = f'Nullable({the_type})'
            return the_type

        result = self.execute(f"DESCRIBE TABLE {resource};", result=True)
        sql = []
        columns = []
        for row in result:
            if row['kind'] == 'COLUMN':
                sql.append(f"{row['name']} {from_snowflake_type(row['type'], row['null?'])}")
                columns.append({
                    'path': row['name'],
                    'name': row['name'],
                    'present_pct': 1 if row['null?'] != YES else None,
                    'recommended_type': from_snowflake_type(row['type'])
                })
        return {
            'analysis': {
                'columns': columns,
                'schema': ', '.join(sql)
            }
        }

    def connector(self):
        auth = {key: self.options[key] for key in self.options.keys() & {'user', 'password', 'account', 'warehouse', 'database', 'schema'}}
        return snowflake.connector.connect(**auth)

    def stage(self):
        return self.options.get('stage') or self.DEFAULT_STAGE

    def storage_integration(self):
        return self.options.get('storage_integration') or self.DEFAULT_STORAGE_INTEGRATION

    def clean(self, blob_name):
        self.gcs.delete(blob_name)


connectors = {
    'snowflake': Snowflake,
    'bigquery': BigQuery,
}


def create_connector(source, params):
    return connectors[source](params)


class DataSensitiveSettings:
    kafka = [
        'kafka_sasl_plain_password'
    ]


class DataConnectorSettings:
    kafka = [
        'cli_version',
        'tb_endpoint',
        'kafka_bootstrap_servers',
        'kafka_sasl_plain_username',
        'kafka_sasl_plain_password',
        'kafka_security_protocol',
        'kafka_sasl_mechanism',
        'kafka_schema_registry_url',
    ]


class DataLinkerSettings:
    kafka = [
        'tb_datasource',
        'tb_token',
        'kafka_topic',
        'kafka_group_id',
        'kafka_auto_offset_reset'
    ]



import asyncio
import json
from typing import Dict, List

import requests
from requests_toolbelt.multipart.encoder import MultipartEncoder
import logging
import ssl
from urllib.parse import quote, urlencode
from tornado.simple_httpclient import SimpleAsyncHTTPClient
from tornado.httpclient import HTTPClientError
from tinybird.syncasync import sync_to_async


HOST = 'https://api.tinybird.co'
LOGIN_URL = 'https://ui.tinybird.co/auth/google'
LIMIT_RETRIES = 10


class AuthException(Exception):
    pass


class AuthNoTokenException(AuthException):
    pass


class DoesNotExistException(Exception):
    pass


class CanNotBeDeletedException(Exception):
    pass


class TimeoutException(Exception):
    pass


class ReachRetryLimit(Exception):
    pass


def connector_equals(connector, datafile_params):
    if not connector:
        return False
    if connector['name'] == datafile_params['kafka_connection_name']:
        return True
    return False


def parse_error_response(response):
    try:
        response = json.loads(response.body)
        return response['error'] if 'error' in response else json.dumps(response, indent=4)
    except json.decoder.JSONDecodeError:
        return 'Server error, cannot parse response'


class TinyB(object):

    def __init__(self, token, host=HOST, version=None):
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

        self.token = token
        self.host = host
        self.version = version
        self._http_client = SimpleAsyncHTTPClient(defaults=dict(allow_nonstandard_methods=True, ssl_options=ctx))

    async def _req(self, endpoint, data=None, method='GET', retries=1, **kwargs):  # noqa: C901
        url = self.host + endpoint

        if self.token:
            url += ('&' if '?' in endpoint else '?') + 'token=' + self.token
        if self.version:
            url += ('&' if '?' in url else '?') + 'cli_version=' + quote(self.version)

        try:
            response = await self._http_client.fetch(url, method=method, body=data, **kwargs)
        except HTTPClientError as e:
            response = e.response
        except Exception as e:
            raise AuthException(f"Error on auth: {e}") from e

        logging.debug("== server response ==")
        logging.debug(response.body)
        logging.debug("==      end        ==")

        if response.code == 403:
            error = parse_error_response(response)
            if not self.token:
                raise AuthNoTokenException(f"Forbidden: {error}")
            raise AuthException(f"Forbidden: {error}")
        if response.code == 204 or response.code == 205:
            return None
        if response.code == 404:
            error = parse_error_response(response)
            raise DoesNotExistException(error)
        if response.code == 500:
            error = parse_error_response(response)
            raise Exception(error)
        if response.code == 409:
            error = parse_error_response(response)
            raise CanNotBeDeletedException(error)
        if response.code == 599:
            raise TimeoutException("timeout")
        if response.code == 429:
            if retries > LIMIT_RETRIES:
                error = parse_error_response(response)
                raise ReachRetryLimit(error)
            retry_after = int(response.headers['Retry-After']) + retries
            await asyncio.sleep(retry_after)
            return await self._req(endpoint, data, method, retries + 1, **kwargs)
        if 'Content-Type' in response.headers and (response.headers['Content-Type'] == 'text/plain' or 'text/csv' in response.headers['Content-Type']):
            return response.body.decode()
        if response.body:
            try:
                response = json.loads(response.body)
            except json.decoder.JSONDecodeError:
                response = {
                    'error': 'Server error, cannot parse response'
                }
        if response.get('error', None):
            error = response['error']
            if response.get('errors', None):
                error += f' -> errors: {response.get("errors")}'
            raise Exception(error)

        return response

    async def tokens(self):
        tokens = await self._req("/v0/tokens")
        return tokens['tokens']

    async def get_token_by_name(self, name):
        tokens = await self.tokens()
        for tk in tokens:
            if tk['name'] == name:
                return tk
        return None

    async def datasources(self, branch=None):
        response = await self._req("/v0/datasources")
        ds = response['datasources']

        if branch:
            ds = [x for x in ds if x['name'].startswith(branch)]
        return ds

    async def connections(self, connector=None):
        response = await self._req('/v0/connectors')
        connectors = response['connectors']

        if connector:
            return [{
                'id': c['id'],
                'service': c['service'],
                'name': c['name'],
                'connected_datasources': len(c['linkers']),
                **c['settings']
            } for c in connectors if c['service'] == connector]
        return [{
            'id': c['id'],
            'service': c['service'],
            'name': c['name'],
            'connected_datasources': len(c['linkers']),
            **c['settings']
        } for c in connectors]

    async def get_datasource(self, ds_name):
        return await self._req(f"/v0/datasources/{ds_name}")

    async def alter_datasource(self, ds_name: str, new_schema: str, dry_run: bool):
        res = await self._req(
            f"/v0/datasources/{ds_name}/alter?dry={'true' if dry_run else 'false'}&schema={quote(new_schema, safe='')}",
            method='POST', data=b'')

        if 'Error' in res:
            raise Exception(res['error'])

        return res

    async def pipe_file(self, pipe):
        return await self._req(f"/v0/pipes/{pipe}.pipe")

    async def datasource_file(self, datasource):
        try:
            return await self._req(f"/v0/datasources/{datasource}.datasource")
        except DoesNotExistException:
            raise Exception(f"Data Source {datasource} not found.")

    async def datasource_analyze(self, url):
        params = {
            'url': url
        }
        return await self._req(f"/v0/analyze?{urlencode(params)}", method='POST')

    async def datasource_analyze_file(self, data):
        return await self._req("/v0/analyze", method='POST', data=data)

    async def datasource_create_from_definition(self, parameter_definition: Dict[str, str]):
        return await self._req(f"/v0/datasources?{urlencode(parameter_definition)}", method='POST', data='')

    async def datasource_create_from_url(self, table_name, url, mode='create', status_callback=None, sql_condition=None, format='csv'):
        params = {
            'name': table_name,
            'url': url,
            'mode': mode,
            'debug': 'blocks_block_log',
            'format': format
        }

        if sql_condition:
            params['replace_condition'] = sql_condition

        req_url = f"/v0/datasources?{urlencode(params, safe='')}"
        res = await self._req(req_url, method='POST', data=b'')

        if 'error' in res:
            raise Exception(res['error'])

        return await self.wait_for_job(res['id'], status_callback, backoff_multiplier=1.5, maximum_backoff_seconds=20)

    async def datasource_delete(self, datasource_name):
        return await self._req(f"/v0/datasources/{datasource_name}", method='DELETE')

    async def datasource_append_data(self, datasource_name, f, mode='append', status_callback=None, sql_condition=None, format='csv'):
        params = {
            'name': datasource_name,
            'mode': mode,
            'format': format,
            'debug': 'blocks_block_log'
        }

        if sql_condition:
            params['replace_condition'] = sql_condition
        if self.version:
            params['cli_version'] = self.version

        url = f"{self.host}/v0/datasources?{urlencode(params, safe='')}"
        if format == 'csv':
            m = MultipartEncoder(fields={'csv': ('csv', f, 'text/csv')})
        else:
            m = MultipartEncoder(fields={'ndjson': ('ndjson', f, 'application/x-ndjson')})

        requests_post = sync_to_async(requests.post)

        r = await requests_post(
            url,
            data=m,
            headers={
                'Authorization': 'Bearer ' + self.token, 'Content-Type': m.content_type
            }
        )

        if r.status_code != 200:
            raise Exception(r.json())

        res = r.json()

        if status_callback:
            status_callback(res)

        return res

    async def datasource_truncate(self, datasource_name):
        return await self._req(f"/v0/datasources/{datasource_name}/truncate", method='POST', data='')

    async def pipes(self, branch=None, dependencies=False, node_attrs=None, attrs=None):
        params = {
            'dependencies': 'true' if dependencies else 'false',
            'attrs': attrs if attrs else '',
            'node_attrs': node_attrs if node_attrs else '',
        }
        response = await self._req(f"/v0/pipes?{urlencode(params)}")
        pipes = response['pipes']
        if branch:
            pipes = [x for x in pipes if x['name'].startswith(branch)]
        return pipes

    async def pipe(self, pipe):
        return await self._req(f"/v0/pipes/{pipe}")

    async def pipe_data(self, pipe_name_or_uid, sql=None):
        if not sql:
            sql = f"SELECT * FROM {pipe_name_or_uid} LIMIT 50"
        return await self._req(f"/v0/pipes/{pipe_name_or_uid}.json?q={quote(sql, safe='')}")

    async def pipe_create(self, pipe_name, sql):
        return await self._req(f"/v0/pipes?name={pipe_name}&sql={quote(sql, safe='')}", method='POST', data=sql.encode())

    async def pipe_delete(self, pipe_name):
        return await self._req(f"/v0/pipes/{pipe_name}", method='DELETE')

    async def pipe_append_node(self, pipe_name_or_uid, sql):
        return await self._req(f"/v0/pipes/{pipe_name_or_uid}/append", method='POST', data=sql.encode())

    async def pipe_set_endpoint(self, pipe_name_or_uid, published_node_uid):
        return await self._req(f"/v0/pipes/{pipe_name_or_uid}/endpoint", method='PUT', data=published_node_uid.encode())

    async def query(self, sql):
        return await self._req(f"/v0/sql?q={quote(sql, safe='')}")

    async def jobs(self, status=None):
        jobs = (await self._req("/v0/jobs"))['jobs']
        if status:
            status = [status] if isinstance(status, str) else status
            jobs = [j for j in jobs if j['status'] in status]
        return jobs

    async def job(self, job_id):
        return await self._req(f"/v0/jobs/{job_id}")

    async def job_cancel(self, job_id):
        return await self._req(f"/v0/jobs/{job_id}/cancel", method='POST', data=b'')

    async def workspaces(self):
        return await self._req("/v0/user/workspaces")

    async def workspace(self, workspace_id, with_token=False):
        with_token = 'true' if with_token else 'false'
        return await self._req(f"/v0/workspaces/{workspace_id}?with_token={with_token}")

    async def workspace_info(self):
        return await self._req("/v0/workspace")

    async def wait_for_job(
            self, job_id, status_callback=None, backoff_seconds=2.0,
            backoff_multiplier: float = 1, maximum_backoff_seconds=2.0):
        done = False

        while not done:
            res = await self._req("/v0/jobs/" + job_id + '?debug=blocks,block_log')

            if res['status'] == 'error':
                error_message = 'There has been an error'
                if not isinstance(res.get('error', True), bool):
                    error_message = str(res['error'])
                if 'errors' in res:
                    error_message += f": {res['errors']}"
                raise Exception(error_message)

            if res['status'] == 'cancelled':
                raise Exception('Job has been cancelled')

            done = res['status'] == 'done'

            if status_callback:
                status_callback(res)

            if not done:
                backoff_seconds = min(backoff_seconds * backoff_multiplier, maximum_backoff_seconds)
                await asyncio.sleep(backoff_seconds)

        return res

    async def datasource_kafka_connect(self, connection_id, datasource_name, topic, group, auto_offset_reset):
        return await self._req(f"/v0/datasources?connector={connection_id}&name={datasource_name}&"
                               f"kafka_topic={topic}&kafka_group_id={group}&kafka_auto_offset_reset={auto_offset_reset}",
                               method='POST', data=b'')

    async def connection_create_kafka(
        self,
        kafka_bootstrap_servers,
        kafka_key,
        kafka_secret,
        kafka_connection_name,
        kafka_auto_offset_reset=None,
        kafka_schema_registry_url=None
    ):

        params = {
            'service': 'kafka',
            'kafka_security_protocol': 'SASL_SSL',
            'kafka_sasl_mechanism': 'PLAIN',
            'kafka_bootstrap_servers': kafka_bootstrap_servers,
            'kafka_sasl_plain_username': kafka_key,
            'kafka_sasl_plain_password': kafka_secret,
            'name': kafka_connection_name
        }

        if kafka_schema_registry_url:
            params['kafka_schema_registry_url'] = kafka_schema_registry_url
        if kafka_auto_offset_reset:
            params['kafka_auto_offset_reset'] = kafka_auto_offset_reset

        connection_params = {
            key: value
            for key, value in params.items() if value is not None
        }

        return await self._req(f"/v0/connectors?{urlencode(connection_params)}", method='POST')

    async def kafka_list_topics(self, connection_id, timeout=5):
        resp = await self._req(f"/v0/connector/preview?connector_id={connection_id}&preview_activity=false",
                               connect_timeout=timeout, request_timeout=timeout)
        return [x['topic'] for x in resp['preview']]

    async def connector_delete(self, connection_id):
        return await self._req(f"/v0/connectors/{connection_id}", method='DELETE')

    async def sql_get_used_tables(self, sql, raising=False) -> List[str]:
        params = {
            'q': sql,
            'raising': 'true' if raising else 'false'
        }
        result = await self._req('/v0/sql_tables', data=urlencode(params), method='POST')
        return [t[1] if t[0] == "" else f"{t[0]}.{t[1]}" for t in result['tables']]

    async def replace_tables(self, q, replacements):
        params = {
            'q': q,
            'replacements': json.dumps({k[1] if isinstance(k, tuple) else k: v
                                        for k, v in replacements.items()})
        }
        result = await self._req('/v0/sql_replace', data=urlencode(params), method='POST')
        return result['query']

    async def get_connection(self, **kwargs):
        result = await self._req('/v0/connectors')
        return next((connector for connector in result['connectors'] if connector_equals(connector, kwargs)), None)

    async def regions(self):
        return await self._req('/v0/regions')



"""
Datafile is like a Dockerfile but to describe ETL processes
"""

import asyncio
import shlex
import sys
import re
import click
import glob
import pprint
import requests
import unittest
from string import Template
from toposort import toposort
from pathlib import Path
import urllib.parse
from urllib.parse import urlencode, urlparse, parse_qs
from collections import namedtuple
from io import StringIO
import os.path
from copy import deepcopy
import traceback

from .sql import parse_table_structure, schema_to_sql_columns
from .client import TinyB, DoesNotExistException
from .sql_template import render_sql_template, get_used_tables_in_template
from .feedback_manager import FeedbackManager
from .ch_utils.engine import ENABLED_ENGINES
from tinybird.syncasync import sync_to_async

INTERNAL_TABLES = ('datasources_ops_log', 'snapshot_views', 'pipe_stats', 'pipe_stats_rt', 'block_log',
                   'data_connectors_log', 'kafka_ops_log')

requests_get = sync_to_async(requests.get, thread_sensitive=False)
requests_post = sync_to_async(requests.post, thread_sensitive=False)
requests_put = sync_to_async(requests.put, thread_sensitive=False)
requests_delete = sync_to_async(requests.delete, thread_sensitive=False)

pp = pprint.PrettyPrinter()


class ParseException(Exception):
    def __init__(self, err, lineno=-1):
        self.lineno = lineno
        super().__init__(err)


class ValidationException(Exception):
    def __init__(self, err, lineno=-1):
        self.lineno = lineno
        super().__init__(err)


def sizeof_fmt(num, suffix='b'):
    """Readable file size
    :param num: Bytes value
    :type num: int
    :param suffix: Unit suffix (optionnal) default = o
    :type suffix: str
    :rtype: str
    """
    for unit in ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def is_shared_datasource(ds_name):
    """just looking for a dot in the name is fine, dot are not allowed in regular datasources"""
    return '.' in ds_name


class Datafile:

    def __init__(self):
        self.maintainer = None
        self.sources = []
        self.nodes = []
        self.tokens = []
        self.keys = []
        self.version = None
        self.description = None

    def validate(self):
        for x in self.nodes:
            if not x['name'].strip():
                raise ValidationException("invalid node name, can't be empty")
            if 'sql' not in x:
                raise ValidationException(
                    "node %s must have a SQL query" % x['name'])
        if self.version is not None and (not isinstance(self.version, int) or self.version < 0):
            raise ValidationException("version must be a positive integer")

    def is_equal(self, other):
        if len(self.nodes) != len(other.nodes):
            return False

        for i, _ in enumerate(self.nodes):
            if self.nodes[i] != other.nodes[i]:
                return False

        return True


def parse_datasource(filename):
    with open(filename) as file:
        s = file.read()
    basepath = os.path.dirname(filename)

    try:
        doc = parse(s, 'default', basepath)
    except ParseException as e:
        raise click.ClickException(FeedbackManager.error_parsing_file(filename=filename, lineno=e.lineno, error=e)) from None

    if len(doc.nodes) > 1:
        raise ValueError(f"{filename}: datasources can't have more than one node")

    return doc


def parse_pipe(filename):
    with open(filename) as file:
        s = file.read()
    basepath = os.path.dirname(filename)

    try:
        doc = parse(s, basepath=basepath)
    except ParseException as e:
        raise click.ClickException(FeedbackManager.error_parsing_file(filename=filename, lineno=e.lineno, error=e))

    return doc


def parse(s, default_node=None, basepath='.'):  # noqa: C901
    """
    Parses `s` string into a document
    >>> d = parse("FROM SCRATCH\\nSOURCE 'https://test.com'\\n#this is a comment\\nMAINTAINER 'rambo' #this is me\\nNODE \\"test_01\\"\\n    DESCRIPTION this is a node that does whatever\\nSQL >\\n\\n        SELECT * from test_00\\n\\n\\nNODE \\"test_02\\"\\n    DESCRIPTION this is a node that does whatever\\nSQL >\\n\\n    SELECT * from test_01\\n    WHERE a > 1\\n   GROUP by a\\n")
    >>> d.maintainer
    'rambo'
    >>> d.sources
    ['https://test.com']
    >>> len(d.nodes)
    2
    >>> d.nodes[0]
    {'name': 'test_01', 'description': 'this is a node that does whatever', 'sql': 'SELECT * from test_00'}
    >>> d.nodes[1]
    {'name': 'test_02', 'description': 'this is a node that does whatever', 'sql': 'SELECT * from test_01\\n    WHERE a > 1\\n   GROUP by a'}
    """
    lines = list(StringIO(s, newline=None))

    doc = Datafile()

    parser_state = namedtuple(
        'ParserState',
        ['multiline', 'current_node', 'command', 'multiline_string'])

    parser_state.multiline = False
    parser_state.current_node = False

    def _unquote(x):
        QUOTES = ('"', "'")
        if x[0] in QUOTES and x[-1] in QUOTES:
            x = x[1:-1]
        return x

    def assign(attr):
        def _fn(x, **kwargs):
            setattr(doc, attr, _unquote(x))
        return _fn

    def schema(*args, **kwargs):
        s = _unquote(''.join(args))
        try:
            sh = parse_table_structure(s)
        except Exception as e:
            raise ParseException(FeedbackManager.error_parsing_schema(line=kwargs['lineno'], error=e))
        parser_state.current_node['schema'] = ','.join(
            schema_to_sql_columns(sh))
        parser_state.current_node['columns'] = sh

    def eval_var(s):
        # replace ENV variables
        # it's probably a bad idea to allow to get any env var
        return Template(s).safe_substitute(os.environ)

    def assign_var(v):
        def _f(*args, **kwargs):
            s = _unquote((' '.join(args)).strip())
            parser_state.current_node[v.lower()] = eval_var(s)
        return _f

    def sources(x, **kwargs):
        doc.sources.append(_unquote(x))

    def node(*args, **kwargs):
        node = {
            'name': eval_var(_unquote(args[0]))
        }
        doc.nodes.append(node)
        parser_state.current_node = node

    def description(*args, **kwargs):
        description = (' '.join(args)).strip()

        if parser_state.current_node:
            parser_state.current_node['description'] = description
        else:
            doc.description = description

    def sql(sql, **kwargs):
        if not parser_state.current_node:
            raise ParseException("SQL must be called after a NODE command")
        parser_state.current_node['sql'] = sql.strip()

    def assign_node_var(v):
        def _f(*args, **kwargs):
            if not parser_state.current_node:
                raise ParseException("%s must be called after a NODE command" % v)
            return assign_var(v)(*args, **kwargs)
        return _f

    def add_token(*args, **kwargs):  # token_name, permissions):
        if len(args) < 2:
            raise ParseException('TOKEN gets two params, token name and permissions e.g TOKEN "read api token" READ')
        doc.tokens.append({
            'token_name': _unquote(args[0]),
            'permissions': args[1]
        })

    def add_key(*args, **kwargs):  # token_name, permissions):
        if len(args) < 1:
            raise ParseException('KEY gets one params')
        doc.keys.append({
            'column': _unquote(args[0])
        })

    def test(*args, **kwargs):
        print("test", args, kwargs)

    def include(*args, **kwargs):
        f = _unquote(args[0])
        f = eval_var(f)
        attrs = dict(_unquote(x).split('=', 1) for x in args[1:])
        nonlocal lines
        lineno = kwargs['lineno']
        # be sure to replace the include line
        p = Path(basepath)
        with open(p / f) as file:
            lines[lineno:lineno + 1] = [''] + list(StringIO(Template(file.read()).safe_substitute(attrs), newline=None))

    def version(*args, **kwargs):
        if len(args) < 1:
            raise ParseException('VERSION gets one positive integer param')
        try:
            version = int(args[0])
            if version < 0:
                raise ValidationException('version must be a positive integer e.g VERSION 2')
            doc.version = version
        except ValueError:
            raise ValidationException('version must be a positive integer e.g VERSION 2')

    def __init_engine(v):
        if not parser_state.current_node:
            raise Exception(f"{v} must be called after a NODE command")
        if 'engine' not in parser_state.current_node:
            parser_state.current_node['engine'] = {'type': None, 'args': []}

    def set_engine(*args, **kwargs):
        __init_engine('ENGINE')
        engine_type = _unquote((' '.join(args)).strip())
        parser_state.current_node['engine']['type'] = engine_type

    def add_engine_var(v):
        def _f(*args, **kwargs):
            __init_engine(f"ENGINE_{v}".upper())
            engine_arg = _unquote((' '.join(args)).strip())
            parser_state.current_node['engine']['args'].append((v, engine_arg))
        return _f

    cmds = {
        'from': assign('from'),
        'source': sources,
        'maintainer': assign('maintainer'),
        'schema': schema,
        'engine_full': assign_var('engine_full'),
        'engine': set_engine,
        'partition_key': assign_var('partition_key'),
        'sorting_key': assign_var('sorting_key'),
        'primary_key': assign_var('primary_key'),
        'sampling_key': assign_var('sampling_key'),
        'ttl': assign_var('ttl'),
        'settings': assign_var('settings'),
        'node': node,
        'description': description,
        'type': assign_node_var('type'),
        'datasource': assign_node_var('datasource'),
        'tags': assign_node_var('tags'),
        'token': add_token,
        'key': add_key,
        'test': test,
        'include': include,
        'sql': sql,
        'version': version,
        'kafka_connection_name': assign_var('kafka_connection_name'),
        'kafka_topic': assign_var('kafka_topic'),
        'kafka_group_id': assign_var('kafka_group_id'),
        'kafka_bootstrap_servers': assign_var('kafka_bootstrap_servers'),
        'kafka_key': assign_var('kafka_key'),
        'kafka_secret': assign_var('kafka_secret'),
        'kafka_schema_registry_url': assign_var('kafka_schema_registry_url'),
        'kafka_target_partitions': assign_var('kafka_target_partitions'),
        'kafka_auto_offset_reset': assign_var('kafka_auto_offset_reset'),
        'kafka_store_raw_value': assign_var('kafka_store_raw_value')
    }

    engine_vars = set()

    for _engine, (params, options) in ENABLED_ENGINES:
        for p in params:
            engine_vars.add(p.name)
        for o in options:
            engine_vars.add(o.name)
    for v in engine_vars:
        cmds[f"engine_{v}"] = add_engine_var(v)

    if default_node:
        node(default_node)

    lineno = 0
    try:
        while lineno < len(lines):
            line = lines[lineno]
            try:
                lexer = list(shlex.shlex(line))
            except ValueError:
                lexer = list(shlex.shlex(shlex.quote(line)))
            if lexer:
                cmd, args = lexer[0], lexer[1:]

                if parser_state.multiline and cmd.lower() in cmds and not (line.startswith(' ') or line.startswith('\t') or line.lower().startswith('from')):
                    parser_state.multiline = False
                    cmds[parser_state.command](parser_state.multiline_string, lineno=lineno)

                if not parser_state.multiline:
                    if len(args) >= 1 and args[0] == ">":
                        parser_state.multiline = True
                        parser_state.command = cmd.lower()
                        parser_state.multiline_string = ''
                    else:
                        if cmd.lower() in cmds:
                            cmds[cmd.lower()](*args, lineno=lineno)
                        else:
                            raise click.ClickException(FeedbackManager.error_option(option=cmd.upper()))
                else:
                    parser_state.multiline_string += line
            lineno += 1
        # close final state
        if parser_state.multiline:
            cmds[parser_state.command](parser_state.multiline_string, lineno=lineno)
    except ParseException as e:
        raise ParseException(str(e), lineno=lineno)
    except ValidationException as e:
        raise ValidationException(str(e), lineno=lineno)
    except IndexError as e:
        raise ValidationException(f'Validation error, found {line} in line {str(lineno)}: {str(e)}', lineno=lineno)
    except Exception as e:
        traceback.print_tb(e.__traceback__)
        raise ParseException(f"Unexpected error: {e}", lineno=lineno)

    return doc


def generate_resource_for_key(key, name, schema, version, res_name):
    resources = []
    # datasource
    ds_name = f"{name}_join_by_{key}{version}"
    params = {
        "name": ds_name,
        "schema": schema,
        "__engine_full": f"Join(ANY, LEFT, {key})"
    }
    resources.append({
        'resource_name': f'{name}_join_by_{key}',
        'resource': 'datasources',
        'params': params,
        'filename': f"{params['name']}.datasource"
    })
    resources.append({
        'resource_name': f'{name}_join_by_{key}_pipe',
        'resource': 'pipes',
        'filename': f"{params['name']}.pipe",
        "name": f"{name}_join_by_{key}_pipe{version}",
        "schema": schema,
        'nodes': [{
            'sql': f"select * from {name}{version}",
            'params': {
                "name": f"{name}_join_by_{key}_view",
                'type': 'materialized',
                'datasource': ds_name
            }
        }],
        'deps': [res_name],  # [ds_name],
        'tokens': []
    })
    return resources


async def process_file(filename, tb_client, dir_path, tag='', resource_versions=None, verbose=False, skip_connectors=False, workspace_map={}, workspace_lib_paths=None):  # noqa: C901 B006

    if resource_versions is None:
        resource_versions = {}
    resource_versions_string = {k: f'__v{v}' for k, v in resource_versions.items() if v >= 0}

    def get_engine_params(node):
        params = {}
        if 'engine_full' in node and 'engine' in node:
            raise ValueError('You can not use ENGINE and ENGINE_FULL in the same node')

        if 'engine_full' in node:
            if verbose:
                click.echo(FeedbackManager.warning_deprecated(warning=f"The 'ENGINE_FULL' keyword is deprecated, found at node '{node.get('name', 'unknown')}', you must use 'ENGINE' instead. 'ENGINE_FULL' will stop working after 2021-03-31."))
            params["__engine_full"] = node['engine_full']

        if 'engine' in node:
            deprecated_vars = ('partition_key', 'sorting_key', 'primary_key', 'sampling_key', 'ttl', 'settings')
            for v in deprecated_vars:
                dep_value = node.get(v, None)
                if dep_value is not None:
                    node['engine']['args'].append((v, dep_value))
                    if verbose:
                        click.echo(FeedbackManager.warning_deprecated_command(keyword=v.upper(), node=node.get('name', 'unknown'), deadline='2021-03-31'))
            engine = node['engine']['type']
            params['engine'] = engine
            args = node['engine']['args']
            for (k, v) in args:
                params[f'engine_{k}'] = v
        return params

    async def get_kafka_params(node):
        params = {
            key: value
            for key, value in node.items() if key.startswith('kafka')
        }

        if not skip_connectors:
            try:
                connector_params = {
                    'kafka_bootstrap_servers': params.get('kafka_bootstrap_servers', None),
                    'kafka_key': params.get('kafka_key', None),
                    'kafka_secret': params.get('kafka_secret', None),
                    'kafka_connection_name': params.get('kafka_connection_name', None),
                    'kafka_auto_offset_reset': params.get('kafka_auto_offset_reset', None),
                    'kafka_schema_registry_url': params.get('kafka_schema_registry_url', None)
                }
                connector = await tb_client.get_connection(**connector_params)

                if not connector:
                    click.echo(FeedbackManager.success_connection_creating(connection_name=params['kafka_connection_name']))
                    connector = await tb_client.connection_create_kafka(**connector_params)
            except Exception as e:
                raise click.ClickException(
                    FeedbackManager.error_connection_create(connection_name=params['kafka_connection_name'], error=str(e)))

            click.echo(FeedbackManager.success_connection_using(connection_name=connector['name']))

            params.update({
                'connector': connector['id'],
                'service': 'kafka',
            })

        return params

    if '.datasource' in filename:
        doc = parse_datasource(filename)
        node = doc.nodes[0]
        deps = []
        # reemplace tables on materialized columns
        columns = parse_table_structure(node['schema'])

        _format = 'csv'
        for x in columns:
            if x['default_value'] and x['default_value'].lower().startswith('materialized'):
                # turn expression to a select query to sql_get_used_tables can get the used tables
                q = 'select ' + x['default_value'][len('materialized'):]
                tables = await tb_client.sql_get_used_tables(q)
                # materualized columns expressions could have joins so we need to add them as a dep
                deps += tables
                # generate replacements and replace the query
                replacements = {t: t + resource_versions_string.get(t, '') for t in tables}

                replaced_results = await tb_client.replace_tables(q, replacements)
                x['default_value'] = replaced_results.replace('SELECT', 'materialized')
            if x.get('jsonpath', x.get('json_path', None)):
                _format = 'ndjson'

        schema = ','.join(schema_to_sql_columns(columns))

        name = os.path.basename(filename).rsplit('.', 1)[0]

        if workspace_lib_paths:
            for wk_name, wk_path in workspace_lib_paths:
                try:
                    Path(filename).relative_to(wk_path)
                    name = f'{workspace_map.get(wk_name, wk_name)}.{name}'
                except ValueError:
                    # the path was not relative, not inside workspace
                    pass
        #
        res_name = name
        if tag and not is_shared_datasource(name):
            name = f'{tag}__{name}'

        version = (f'__v{doc.version}' if doc.version is not None else '')

        def append_version_to_name(name, version):
            if version != '':
                name = name.replace(".", "_")
                return name + version
            return name

        params = {
            "name": append_version_to_name(name, version),
            "schema": schema,
            "format": _format
        }

        params.update(get_engine_params(node))

        if 'kafka_connection_name' in node:
            kafka_params = await get_kafka_params(node)
            params.update(kafka_params)
            # drop the schema, it's fixed by the kafka connector
            del params["schema"]
            del params["engine"]
            del params["format"]

        if 'tags' in node:
            tags = {k: v[0]
                    for k, v in urllib.parse.parse_qs(node['tags']).items()}
            params.update(tags)

        resources = []

        resources.append({
            'resource': 'datasources',
            'resource_name': name,
            "version": doc.version,
            'params': params,
            'filename': os.path.basename(filename),
            'keys': doc.keys,
            'deps': deps
        })

        # generate extra resources in case of the key
        if doc.keys:
            for k in doc.keys:
                resources += generate_resource_for_key(
                    k['column'],
                    name,
                    params['schema'],
                    version,
                    res_name
                )
                # set derived resources version the same the parent
                for x in resources:
                    x['version'] = doc.version

        return resources

    elif '.pipe' in filename:
        doc = parse_pipe(filename)
        version = (f'__v{doc.version}' if doc.version is not None else '')
        name = os.path.basename(filename).split('.')[0]
        description = doc.description if doc.description is not None else ''

        if tag and not is_shared_datasource(name):
            name = f'{tag}__{name}'

        deps = []
        nodes = []

        for node in doc.nodes:
            sql = node['sql']
            params = {
                'name': node['name'],
                'type': node.get('type', 'standard'),
                'description': node.get('description', ''),
            }
            if node.get('type', '').lower() == 'materialized':
                params.update({
                    'type': 'materialized',
                })

            sql = sql.strip()
            is_template = False
            if sql[0] == '%':
                sql_rendered = render_sql_template(sql[1:], test_mode=True)
                is_template = True
            else:
                sql_rendered = sql

            try:
                dependencies = await tb_client.sql_get_used_tables(sql_rendered, raising=True)
                deps += [t for t in dependencies if t not in [n['name'] for n in doc.nodes]]

            except Exception as e:
                raise click.ClickException(FeedbackManager.error_parsing_node(
                    node=node['name'], pipe=name, error=str(e)))

            if is_template:
                deps += get_used_tables_in_template(sql[1:])

            tag_ = ''
            if tag:
                tag_ = f'{tag}__'

            if ('engine_full' in node or 'engine' in node) and 'datasource' not in node:
                raise ValueError('Defining ENGINE options in a node requires a DATASOURCE')

            if 'datasource' in node:
                params['datasource'] = tag_ + node['datasource'] + \
                    resource_versions_string.get(tag_ + node['datasource'], '')
                deps += [node['datasource']]

            params.update(get_engine_params(node))

            def create_replacement_for_resource(tag, name):
                for old_ws, new_ws in workspace_map.items():
                    name = name.replace(f'{old_ws}.', f'{new_ws}.')
                if tag != '' and not is_shared_datasource(name):
                    name = tag + name
                return name + resource_versions_string.get(name, '')

            replacements = {x: create_replacement_for_resource(tag_, x) for x in deps if x not in [
                n['name'] for n in doc.nodes]}

            # when using templates replace_tables from clickhouse can't be used. Not ideal but ok
            if is_template:
                for old, new in replacements.items():
                    sql = re.sub(
                        '([\t \\n\']+|^)' + old + '([\t \\n\'\\)]+|$)', "\\1" + new + "\\2", sql)
            else:
                sql = await tb_client.replace_tables(sql, replacements)

            if 'tags' in node:
                tags = {k: v[0]
                        for k, v in urllib.parse.parse_qs(node['tags']).items()}
                params.update(tags)

            nodes.append({
                'sql': sql,
                'params': params
            })

        return [{
            'resource': 'pipes',
            'resource_name': name,
            "version": doc.version,
            'filename': os.path.basename(filename),
            'name': name + version,
            'nodes': nodes,
            'deps': [x for x in set(deps)],
            'tokens': doc.tokens,
            'description': description
        }]
    else:
        raise Exception(FeedbackManager.error_file_extension(filename=filename))


def full_path_by_name(folder, name, workspace_lib_paths=None):
    f = Path(folder)
    ds = name + ".datasource"
    if os.path.isfile(os.path.join(folder, ds)):
        return f / ds
    if os.path.isfile(f / 'datasources' / ds):
        return f / 'datasources' / ds

    pipe = name + ".pipe"
    if os.path.isfile(os.path.join(folder, pipe)):
        return f / pipe

    if os.path.isfile(f / 'endpoints' / pipe):
        return f / 'endpoints' / pipe

    if os.path.isfile(f / 'pipes' / pipe):
        return f / 'pipes' / pipe

    if workspace_lib_paths:
        for wk_name, wk_path in workspace_lib_paths:
            if name.startswith(f'{wk_name}.'):
                r = full_path_by_name(wk_path, name.replace(f'{wk_name}.', ''))
                if r:
                    return r


def find_file_by_name(folder, name, verbose=False, is_raw=False, workspace_lib_paths=None):
    f = Path(folder)
    ds = name + ".datasource"
    if os.path.isfile(os.path.join(folder, ds)):
        return ds
    if os.path.isfile(f / 'datasources' / ds):
        return ds

    pipe = name + ".pipe"
    if os.path.isfile(os.path.join(folder, pipe)):
        return pipe

    if os.path.isfile(f / 'endpoints' / pipe):
        return pipe

    if os.path.isfile(f / 'pipes' / pipe):
        return pipe

    # look for the file in subdirectories if it's not found in datasources folder
    if workspace_lib_paths:
        r = None
        for wk_name, wk_path in workspace_lib_paths:
            if name.startswith(f'{wk_name}.'):
                r = find_file_by_name(wk_path, name.replace(f'{wk_name}.', ''), verbose, is_raw)
            if r:
                return r

    if not is_raw:
        return find_file_by_name(folder, get_dep_from_raw_tables(name), verbose=verbose, is_raw=True, workspace_lib_paths=workspace_lib_paths)

    if verbose:
        click.echo(FeedbackManager.warning_file_not_found_inside(name=name, folder=folder))


def drop_token(url):
    """
    drops token param from the url query string
    >>> drop_token('https://api.tinybird.co/v0/pipes/aaa.json?token=abcd&a=1')
    'https://api.tinybird.co/v0/pipes/aaa.json?a=1'
    >>> drop_token('https://api.tinybird.co/v0/pipes/aaa.json?a=1')
    'https://api.tinybird.co/v0/pipes/aaa.json?a=1'
    """
    parsed = urlparse(url)
    qs = parse_qs(parsed.query)
    qs = {k: v[0] for k, v in qs.items()}  # change several arguments to single one
    if 'token' in qs:
        del qs['token']
    return f"{parsed.scheme}://{parsed.netloc}{parsed.path}?{urlencode(qs)}"


class PipeChecker(unittest.TestCase):
    def __init__(self, current_pipe_url, checker_pipe_name, token, only_response_times, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_pipe_url = drop_token(current_pipe_url)
        self.checker_pipe_name = checker_pipe_name
        self.token = token
        self.only_response_times = only_response_times
        parsed = urlparse(self.current_pipe_url)
        self.qs = parse_qs(parsed.query)
        self.checker_pipe_url = f"{parsed.scheme}://{parsed.netloc}/v0/pipes/{self.checker_pipe_name}.json?{parsed.query}"

    def __str__(self):
        return f"current {self.current_pipe_url}\n    new {self.checker_pipe_url}"

    def runTest(self):
        if 'debug' in self.qs:
            self.skipTest('found debug param')
        headers = {'Authorization': f'Bearer {self.token}'}

        current_r = requests.get(self.current_pipe_url, headers=headers)
        checker_r = requests.get(self.checker_pipe_url, headers=headers)
        current_data = current_r.json().get('data', [])
        check_data = checker_r.json().get('data', [])
        error_check_data = checker_r.json().get('error', None)
        self.assertIsNone(error_check_data, 'You are trying to push a pipe with errors, please check the output or run with --no-check')
        if not self.only_response_times:
            self.assertEqual(len(current_data), len(check_data), "Number of elements does not match")
        for i, (current_data_e, check_data_e) in enumerate(zip(current_data, check_data)):
            if self.only_response_times:
                self.assertAlmostEqual(
                    current_r.elapsed.total_seconds(),
                    checker_r.elapsed.total_seconds(),
                    delta=.25,
                    msg="response time has changed by more than 25%"
                )
                continue
            self.assertEqual(list(current_data_e.keys()),
                             list(check_data_e.keys()))
            for x in current_data_e.keys():
                if type(current_data_e[x]) == float:
                    d = abs(current_data_e[x] - check_data_e[x])
                    self.assertLessEqual(d / current_data_e[x], 0.001)
                else:
                    self.assertEqual(
                        current_data_e[x], check_data_e[x], f"Failed on index {i}, key {x}")


async def check_pipe(pipe, host, token, populate, cl, limit=10, only_response_times=False):
    checker_pipe = deepcopy(pipe)
    checker_pipe['name'] = f"{checker_pipe['name']}__checker"
    if populate:
        raise Exception(FeedbackManager.error_check_pipes_populate())

    await new_pipe(checker_pipe, cl, replace=True, check=False, populate=populate, skip_tokens=True, skip_table_checks=True, ignore_sql_errors=True)
    headers = {'Authorization': f'Bearer {token}'}
    r = await requests_get(f"{host}/v0/pipes/{pipe['name']}/requests", headers=headers)
    pipe_top_requests = r.json().get('requests', {}).get('top', [])
    suite = unittest.TestSuite()

    for i, r in enumerate(pipe_top_requests):
        if i == limit:
            break
        suite.addTest(PipeChecker(r['endpoint_url'],
                                  checker_pipe['name'], token, only_response_times))
    result = unittest.TextTestResult(unittest.runner._WritelnDecorator(sys.stdout), descriptions=True, verbosity=2)
    suite.run(result)

    if not result.wasSuccessful():
        for _test, err in result.failures:
            try:
                i = err.index('AssertionError') + len('AssertionError :')
                click.echo('==== Test FAILED ====\n')
                click.echo(_test)
                click.echo(FeedbackManager.error_check_pipe(error=err[i:]))
                click.echo('=====================\n\n\n')
            except Exception:
                pass
        raise RuntimeError('Invalid results, you can bypass checks by running push with the --no-check flag')

    # Only delete if no errors, so we can check results after failure
    r = await requests_delete(f"{host}/v0/pipes/{checker_pipe['name']}", headers=headers)
    if r.status_code != 204:
        click.echo(FeedbackManager.warning_check_pipe(content=r.content))


async def new_pipe(p, tb_client: TinyB, replace=False, check=True, populate=False, wait_populate=False, skip_tokens=False, tag='', skip_table_checks=False, ignore_sql_errors=False, only_response_times=False, timeout=None):  # noqa: C901
    exists = False
    materialized = False

    # TODO use tb_client instead of calling the urls directly.
    host = tb_client.host
    token = tb_client.token

    headers = {'Authorization': f'Bearer {token}'}

    if tag:
        tag = tag + "__"

    cli_params = {}
    cli_params['cli_version'] = tb_client.version
    cli_params['description'] = p.get('description', '')
    cli_params['skip_table_checks'] = 'true' if skip_table_checks else 'false'
    cli_params['ignore_sql_errors'] = 'true' if ignore_sql_errors else 'false'

    r = await requests_get(f"{host}/v0/pipes/{p['name']}?{urlencode(cli_params)}", headers=headers)

    if r.status_code == 200:
        exists = True
        if replace:
            # TODO: this should create a different node and rename it to the final one on success
            if check and not populate:
                await check_pipe(p, host, token, populate, tb_client, only_response_times=only_response_times)

            pipe = r.json()

            delete_endpoint = await requests_put(f"{host}/v0/pipes/{p['name']}/endpoint?{urlencode(cli_params)}", headers=headers, data='')
            if delete_endpoint.status_code != 200:
                raise Exception(FeedbackManager.error_remove_endpoint(error=r.content))

            update_pipe = await requests_put(f"{host}/v0/pipes/{p['name']}?{urlencode(cli_params)}", headers=headers, data='')
            if update_pipe.status_code != 200:
                raise Exception(FeedbackManager.error_updating_pipe(error=r.content))

            for n in pipe.get('nodes', []):
                r = await requests_delete(f"{host}/v0/pipes/{p['name']}/nodes/{n['id']}?{urlencode(cli_params)}", headers=headers)
                if r.status_code != 204:
                    raise Exception(FeedbackManager.error_removing_node(pipe=p['name'], error=r.json()['error']))
        else:
            raise click.ClickException(FeedbackManager.error_pipe_already_exists(pipe=p['name']))
    else:
        r = await requests_post(f"{host}/v0/pipes?name={p['name']}&sql=select+1&{urlencode(cli_params)}", headers=headers, data='')
        if r.status_code != 200:
            try:
                raise Exception(FeedbackManager.error_creating_pipe(error=r.json()['error']))
            except ValueError:
                raise Exception(FeedbackManager.error_creating_pipe(error=r.content))

        dummy_node = r.json()['nodes'][0]['id']

        r = await requests_delete(f"{host}/v0/pipes/{p['name']}/nodes/{dummy_node}?{urlencode(cli_params)}", headers=headers)
        if r.status_code != 204:
            raise Exception(FeedbackManager.error_removing_dummy_node(error=r.content))

    for node in p['nodes']:
        params = node['params']
        if populate:
            params['populate'] = 'true'
        if params.get('type', '') == 'materialized':
            materialized = True
        params['cli_version'] = tb_client.version
        params['skip_table_checks'] = 'true' if skip_table_checks else 'false'
        params['ignore_sql_errors'] = 'true' if ignore_sql_errors else 'false'

        r = await requests_post(f"{host}/v0/pipes/{p['name']}/nodes?{urlencode(params)}", headers=headers, data=node['sql'])

        if r.status_code != 200:
            if not exists:
                # remove the pipe when it does not exist
                await requests_delete(f"{host}/v0/pipes/{p['name']}?{urlencode(cli_params)}", headers=headers)
            raise Exception(FeedbackManager.error_creating_node(pipe=p['name'], error=r.json()['error']))

        new_node = r.json()
        pipe_type = params.get('type', None)

        if pipe_type == 'materialized':
            created_datasource = new_node.get('created_datasource', False)
            datasource = new_node.get('datasource')

            if created_datasource:
                click.echo(FeedbackManager.info_materialized_datasource_created(node=new_node.get('name', 'unknown'), datasource=datasource['name']))
            else:
                click.echo(FeedbackManager.info_materialized_datasource_used(node=new_node.get('name', 'unknown'), datasource=datasource['name']))

            if populate:
                job_url = new_node.get('job', {}).get('job_url', None)
                job_id = new_node.get('job', {}).get('job_id', None)
                click.echo(FeedbackManager.info_populate_job_url(url=job_url))

                if wait_populate:
                    with click.progressbar(label="Populating ", length=100, show_eta=False, show_percent=True, fill_char=click.style("█", fg="green")) as progress_bar:
                        def progressbar_cb(res):
                            if 'progress_percentage' in res:
                                progress_bar.update(int(round(res['progress_percentage'])) - progress_bar.pos)
                            elif res['status'] != 'working':
                                progress_bar.update(progress_bar.length)
                        try:
                            result = await asyncio.wait_for(tb_client.wait_for_job(job_id, status_callback=progressbar_cb), timeout)
                            if result['status'] != 'done':
                                click.echo(FeedbackManager.error_while_populating(error=result['error']))
                        except asyncio.TimeoutError:
                            await tb_client.job_cancel(job_id)
                            raise click.ClickException(FeedbackManager.error_while_populating(error="Reach timeout, job cancelled"))
                        except Exception as e:
                            raise click.ClickException(FeedbackManager.error_while_populating(error=str(e)))

        endpoint_node = r.json()['id']
        r = await requests_put(f"{host}/v0/pipes/{p['name']}/endpoint?{urlencode(cli_params)}", headers=headers, data=endpoint_node)
        if r.status_code != 200:
            raise Exception(FeedbackManager.error_creating_endpoint(node=endpoint_node, pipe=p['name'], error=r.json()['error']))

    if p['tokens'] and not skip_tokens:

        # search for token with specified name and adds it if not found or adds permissions to it
        t = None
        for tk in p['tokens']:
            token_name = tag + tk['token_name']
            t = await tb_client.get_token_by_name(token_name)
            if t:
                break
        if not t:
            token_name = tag + tk['token_name']
            click.echo(FeedbackManager.info_create_not_found_token(token=token_name))
            url = f"{host}/v0/tokens?name={token_name}&scope=PIPES:{tk['permissions']}:{p['name']}&{urlencode(cli_params)}"
            r = await requests_post(url, headers=headers, data='')
            if r.status_code != 200:
                raise Exception(FeedbackManager.error_creating_pipe(error=r.json()['error']))
            token = r.json()['token']
        else:
            click.echo(FeedbackManager.info_create_found_token(token=token_name))
            scopes = [f"PIPES:{tk['permissions']}:{p['name']}"]
            for x in t['scopes']:
                scopes.append(f"{x['type']}:{x['resource']}")
            s = '&'.join([f'scope={x}' for x in scopes])
            url = f"{host}/v0/tokens/{t['token']}?{s}&{urlencode(cli_params)}"
            r = await requests_put(url, headers=headers, data='')
            if r.status_code != 200:
                raise Exception(FeedbackManager.error_creating_pipe(error=r.json()['error']))
            token = t['token']
        click.echo(FeedbackManager.success_test_endpoint(host=host, pipe=p['name'], token=token))
    else:
        if not materialized and not skip_tokens:
            click.echo(FeedbackManager.success_test_endpoint_no_token(host=host, pipe=p['name']))


async def new_ds(ds, client, replace=False, skip_table_checks=False, skip_confirmation=False):
    ds_name = ds['params']['name']

    try:
        await client.get_datasource(ds_name)
        datasource_exists = True
    except DoesNotExistException:
        datasource_exists = False

    if not datasource_exists:
        params = ds['params']
        params['skip_table_checks'] = 'true' if skip_table_checks else 'false'

        try:
            await client.datasource_create_from_definition(params)
        except Exception as e:
            raise Exception(FeedbackManager.error_creating_datasource(error=str(e)))
        click.echo(FeedbackManager.success_create(name=ds["resource_name"]))
        return

    if not replace:
        raise click.ClickException(FeedbackManager.error_datasource_already_exists(datasource=ds_name))

    alter_response = None
    alter_error_message = None

    try:
        # Schema fixed by the kafka connector
        if 'kafka_connection_name' not in ds['params']:
            alter_response = await client.alter_datasource(ds_name, ds['params']['schema'], True)
    except Exception as e:
        if "There were no operations to perform" in str(e):
            pass
        else:
            alter_error_message = str(e)

    if alter_response:
        click.echo(FeedbackManager.info_datasource_doesnt_match_current_schema(datasource=ds_name))
        for operation in alter_response["operations"]:
            click.echo(f"**   -  {operation}")

        if skip_confirmation:
            make_changes = True
        else:
            make_changes = click.prompt(FeedbackManager.info_ask_for_alter_confirmation()).lower() == "y"

        if make_changes:
            await client.alter_datasource(ds_name, ds['params']['schema'], False)
            click.echo(FeedbackManager.success_datasource_alter())
            return
        else:
            alter_error_message = "Alter datasource cancelled"

    # removed replacing by default. When a datasource is removed data is
    # removed and all the references needs to be updated
    if os.getenv('TB_I_KNOW_WHAT_I_AM_DOING') and click.prompt(FeedbackManager.info_ask_for_datasource_confirmation()) == ds_name:  # TODO move to CLI
        try:
            await client.datasource_delete(ds_name)
            click.echo(FeedbackManager.success_delete_datasource(datasource=ds_name))
        except Exception:
            raise Exception(FeedbackManager.error_removing_datasource(datasource=ds_name))
        return
    else:
        if alter_error_message:
            raise click.ClickException(FeedbackManager.error_datasource_already_exists_and_alter_failed(
                datasource=ds_name, alter_error_message=alter_error_message))
        else:
            click.echo(FeedbackManager.warning_datasource_already_exists(datasource=ds_name))


async def exec_file(r, tb_client, override, check, debug, populate, wait_populate, tag, skip_table_checks=False, ignore_sql_errors=False, skip_confirmation=False, only_response_times=False, timeout=None):

    if debug:
        click.echo(FeedbackManager.debug_running_file(file=pp.pformat(r)))
    if r['resource'] == 'pipes':
        await new_pipe(
            r,
            tb_client,
            override,
            check,
            populate,
            wait_populate,
            tag=tag,
            skip_table_checks=skip_table_checks,
            ignore_sql_errors=ignore_sql_errors,
            only_response_times=only_response_times,
            timeout=timeout)
        click.echo(FeedbackManager.success_create(name=r["resource_name"]))

    elif r['resource'] == 'datasources':
        await new_ds(
            r,
            tb_client,
            override,
            skip_table_checks=skip_table_checks,
            skip_confirmation=skip_confirmation
        )
    else:
        raise Exception(FeedbackManager.error_unknown_resource(resource=r['resource']))


def get_name_tag_version(ds):
    """
    Given a name like "name__dev__v0" returns ['name', 'dev', 'v0']
    >>> get_name_tag_version('dev__name__v0')
    {'name': 'name', 'tag': 'dev', 'version': 0}
    >>> get_name_tag_version('name__v0')
    {'name': 'name', 'tag': None, 'version': 0}
    >>> get_name_tag_version('dev__name')
    {'name': 'name', 'tag': 'dev', 'version': None}
    >>> get_name_tag_version('name')
    {'name': 'name', 'tag': None, 'version': None}
    >>> get_name_tag_version('horario__3__pipe')
    {'name': '3__pipe', 'tag': 'horario', 'version': None}
    >>> get_name_tag_version('horario__checker')
    {'name': 'horario__checker', 'tag': None, 'version': None}
    >>> get_name_tag_version('tg__dActividades__v0_pipe_3907')
    {'name': 'dActividades', 'tag': 'tg', 'version': 0}
    >>> get_name_tag_version('tg__dActividades__va_pipe_3907')
    {'name': 'dActividades__va_pipe_3907', 'tag': 'tg', 'version': None}
    >>> get_name_tag_version('tg__origin_workspace.shared_ds__v3907')
    {'name': 'origin_workspace.shared_ds', 'tag': 'tg', 'version': 3907}
    """
    tk = ds.rsplit('__', 2)
    if len(tk) == 1:
        return {'name': tk[0], 'tag': None, 'version': None}
    elif len(tk) == 2:
        if tk[1][0] == 'v' and re.match('[0-9]+$', tk[1][1:]):
            return {'name': tk[0], 'tag': None, 'version': int(tk[1][1:])}
        else:
            if tk[1] == 'checker':
                return {'name': tk[0] + "__" + tk[1], 'tag': None, 'version': None}
            return {'name': tk[1], 'tag': tk[0], 'version': None}
    elif len(tk) == 3:
        if tk[2] == 'checker':
            return {'name': tk[1], 'tag': tk[0], 'version': None}
        if tk[2][0] == 'v':
            parts = tk[2].split('_')
            try:
                return {'name': tk[1], 'tag': tk[0], 'version': int(parts[0][1:])}
            except ValueError:
                return {'name': f'{tk[1]}__{tk[2]}', 'tag': tk[0], 'version': None}
        else:
            return {'name': '__'.join(tk[1:]), 'tag': tk[0], 'version': None}

    return ds


def get_resource_versions(datasources):
    """
    return the latest version for all the datasources
    """
    versions = {}
    for x in datasources:
        t = get_name_tag_version(x)
        name = t['name']
        if t['tag']:
            name = f"{t['tag']}__{name}"
        if t.get('version', None) is not None:
            versions[name] = t['version']
    return versions


def get_remote_resource_name_without_version(remote_resource_name: str) -> str:
    """
    >>> get_remote_resource_name_without_version("r__datasource")
    'r__datasource'
    >>> get_remote_resource_name_without_version("r__datasource__v0")
    'r__datasource'
    >>> get_remote_resource_name_without_version("datasource")
    'datasource'
    """
    parts = get_name_tag_version(remote_resource_name)
    if parts['tag']:
        return parts['tag'] + '__' + parts['name']
    else:
        return parts['name']


def get_dep_from_raw_tables(x):
    """
    datasources KEY command generates tables, this transform the used table with the source file
    >>> get_dep_from_raw_tables('test')
    'test'
    >>> get_dep_from_raw_tables('test_join_by_column')
    'test'
    """

    try:
        return x[:x.index('_join_by_')]
    except ValueError:
        return x


async def build_graph(filenames, tb_client, tag='', resource_versions=None, workspace_map=None, process_dependencies=False, verbose=False, skip_connectors=False, workspace_lib_paths=None):  # noqa: C901 B006
    """process files"""
    to_run = {}
    deps = []
    dep_map = {}
    if not workspace_map:
        workspace_map = {}

    deps_tag = ''
    if tag:
        deps_tag = f'{tag}__'

    dir_path = os.getcwd()

    async def process(filename, deps, dep_map, to_run, workspace_lib_paths):
        name, kind = filename.rsplit('.', 1)

        try:
            res = await process_file(
                filename,
                tb_client,
                dir_path,
                tag,
                resource_versions=resource_versions,
                verbose=verbose,
                skip_connectors=skip_connectors,
                workspace_map=workspace_map,
                workspace_lib_paths=workspace_lib_paths
            )
        except click.ClickException as e:
            raise e
        except Exception as e:
            raise click.ClickException(e)

        for r in res:
            fn = r['resource_name']
            to_run[fn] = r
            file_deps = r.get('deps', [])
            deps += file_deps
            # calculate and look for deps
            dep_list = []
            for x in file_deps:
                if x not in INTERNAL_TABLES:
                    f = find_file_by_name(dir_path, x, verbose, workspace_lib_paths=workspace_lib_paths)
                    if f:
                        dep_list.append(deps_tag + f.rsplit('.', 1)[0])
            dep_map[fn] = set(dep_list)
        return os.path.basename(name)

    processed = set()
    for filename in filenames:
        if verbose:
            click.echo(FeedbackManager.info_processing_file(filename=filename))
        name = await process(filename, deps, dep_map, to_run, workspace_lib_paths)
        processed.add(name)

    if process_dependencies:
        while len(deps) > 0:
            dep = deps.pop()
            if dep not in processed:
                processed.add(dep)
                f = full_path_by_name(dir_path, dep, workspace_lib_paths)
                if f:
                    if verbose:
                        try:
                            processed_filename = f.relative_to(os.getcwd())
                        except ValueError:
                            processed_filename = f
                        click.echo(FeedbackManager.info_processing_file(filename=processed_filename))
                    await process(str(f), deps, dep_map, to_run, workspace_lib_paths)

    return to_run, dep_map


def get_project_filenames(folder):
    folders = [
        f'{folder}/*.datasource',
        f'{folder}/datasources/*.datasource',
        f'{folder}/*.pipe',
        f'{folder}/pipes/*.pipe',
        f'{folder}/endpoints/*.pipe',
    ]
    filenames = []
    for x in folders:
        filenames += glob.glob(x)
    return filenames


async def folder_push(tb_client: TinyB, tag='', filenames=None, dry_run=False, check=False, push_deps=False,  # noqa: C901
                      debug=False, force=False, folder='.', populate=False, upload_fixtures=False, wait=False,
                      skip_table_checks=False, ignore_sql_errors=False, skip_confirmation=None, only_response_times=False,
                      workspace_map=None, workspace_lib_paths=None, no_versions=False, timeout=None):

    if not workspace_map:
        workspace_map = {}
    if not workspace_lib_paths:
        workspace_lib_paths = []

    workspace_lib_paths = list(workspace_lib_paths)
    # include vendor libs without overriding user ones without overriding user ones
    existing_workspaces = set(x[1] for x in workspace_lib_paths)
    vendor_path = Path('vendor')
    if vendor_path.exists():
        for x in vendor_path.iterdir():
            if x.is_dir() and x.name not in existing_workspaces:
                workspace_lib_paths.append((x.name, x))

    datasources = await tb_client.datasources()
    pipes = await tb_client.pipes()

    existing_resources = [x['name'] for x in datasources] + [x['name'] for x in pipes]
    # replace workspace mapping names
    for old_ws, new_ws in workspace_map.items():
        existing_resources = [x.replace(f'{old_ws}.', f'{new_ws}.') for x in existing_resources]

    if not no_versions:
        resource_versions = get_resource_versions(existing_resources)
    else:
        resource_versions = {}

    remote_resource_names = [get_remote_resource_name_without_version(x) for x in existing_resources]

    # replace workspace mapping names
    for old_ws, new_ws in workspace_map.items():
        remote_resource_names = [x.replace(f'{old_ws}.', f'{new_ws}.') for x in remote_resource_names]

    if not filenames:
        filenames = get_project_filenames(folder)

    # build graph to get new versions for all the files involved in the query
    # dependencies need to be processed always to get the versions
    resources, dep_map = await build_graph(filenames, tb_client, tag=tag, process_dependencies=True, workspace_map=workspace_map, skip_connectors=True, workspace_lib_paths=workspace_lib_paths)

    # update existing versions
    if not no_versions:
        latest_datasource_versions = resource_versions.copy()

        for dep in resources.values():
            ds = dep['resource_name']
            if dep['version'] is not None:
                latest_datasource_versions[ds] = dep['version']
    else:
        latest_datasource_versions = {}

    # build the graph again with the rigth version
    to_run, dep_map = await build_graph(
        filenames,
        tb_client,
        tag=tag,
        resource_versions=latest_datasource_versions,
        workspace_map=workspace_map,
        process_dependencies=push_deps,
        verbose=True,
        workspace_lib_paths=workspace_lib_paths
    )

    if debug:
        pp.pprint(to_run)

    click.echo(FeedbackManager.info_building_dependencies())

    for group in toposort(dep_map):
        for name in group:
            if name in to_run:
                if not dry_run:
                    if name not in remote_resource_names or resource_versions.get(name.replace(".", "_"), '') != latest_datasource_versions.get(name, '') or force:
                        if name not in resource_versions:
                            version = ''
                            if name in latest_datasource_versions:
                                version = f'(v{latest_datasource_versions[name]})'
                            click.echo(FeedbackManager.info_processing_new_resource(name=name, version=version))
                        else:
                            click.echo(FeedbackManager.info_processing_resource(
                                name=name,
                                version=latest_datasource_versions[name],
                                latest_version=resource_versions.get(name)
                            ))
                        try:
                            await exec_file(to_run[name], tb_client, force, check, debug, populate, wait, tag,
                                            skip_table_checks, ignore_sql_errors, skip_confirmation, only_response_times, timeout)
                        except Exception as e:
                            exception = FeedbackManager.error_push_file_exception(filename=to_run[name]['filename'], error=e)
                            if '[CLI] skip_table_checks_error' in exception:
                                skip_table_checks_error = FeedbackManager.error_skip_table_checks_path(path=to_run[name]['filename'])
                                exception = exception.replace('[CLI] skip_table_checks_error', f', {skip_table_checks_error}')
                            raise click.ClickException(exception)
                    else:
                        click.echo(FeedbackManager.warning_name_already_exists(name=name))
                else:
                    if name not in remote_resource_names or resource_versions.get(name, '') != latest_datasource_versions.get(name, '') or force:
                        click.echo(FeedbackManager.info_dry_creating_resource(name=name))
                    else:
                        click.echo(FeedbackManager.warning_dry_name_already_exists(name=name))

    if not dry_run:
        if upload_fixtures:
            click.echo(FeedbackManager.info_pushing_fixtures())
            processed = set()
            for group in toposort(dep_map):
                for f in group:
                    name = os.path.basename(f)
                    if name not in processed:
                        if name in to_run:
                            await check_data(tb_client, to_run[name], debug, folder)
                            processed.add(name)
            for f in to_run:
                if f not in processed:
                    await check_data(tb_client, to_run[f], debug, folder)
        else:
            click.echo(FeedbackManager.info_not_pushing_fixtures())


async def check_data(cl, r, debug, folder=''):
    if debug:
        click.echo(FeedbackManager.info_checking_file(file=pp.pformat(r)))
    if r['resource'] == 'pipes':
        pass
    elif r['resource'] == 'datasources':
        name = os.path.basename(r['filename']).rsplit('.', 1)[0]
        csv_test_file = Path(folder) / 'fixtures' / f'{name}.csv'
        if not csv_test_file.exists():
            csv_test_file = Path(folder) / 'datasources' / 'fixtures' / f'{name}.csv'
        if csv_test_file.exists():
            click.echo(FeedbackManager.info_checking_file_size(filename=r['filename'], size=sizeof_fmt(os.stat(csv_test_file).st_size)))
            sys.stdout.flush()
            try:
                with open(csv_test_file, 'rb') as file:
                    res = await cl.datasource_append_data(r['params']['name'], file)
                click.echo(FeedbackManager.success_processing_blocks(num_blocks=len(res['blocks'])))
            except Exception as e:
                raise click.ClickException(FeedbackManager.error_processing_blocks(error=e))

        else:
            click.echo(FeedbackManager.warning_file_not_found(name=csv_test_file))

    else:
        raise Exception(FeedbackManager.error_unknown_resource(resource=r['resource']))



import re
from collections import defaultdict

from ..sql import col_name, engine_replicated_to_local, parse_table_structure


class TableDetails:
    '''
    >>> ed = TableDetails({})
    >>> ed.engine_full == None
    True
    >>> ed.engine == None
    True
    >>> ed.to_json()
    {'engine_full': None, 'engine': None}
    >>> ed.to_datafile()
    ''

    >>> ed = TableDetails({ "engine_full": "MergeTree() PARTITION BY toYear(timestamp) ORDER BY (timestamp, cityHash64(location)) SAMPLE BY cityHash64(location) SETTINGS index_granularity = 32, index_granularity_bytes = 2048 TTL toDate(timestamp) + INTERVAL 1 DAY", "engine": "MergeTree", "partition_key": "toYear(timestamp)", "sorting_key": "timestamp, cityHash64(location)", "primary_key": "timestamp, cityHash64(location)", "sampling_key": "cityHash64(location)", "settings": "index_granularity = 32, index_granularity_bytes = 2048", "ttl": "toDate(timestamp) + INTERVAL 1 DAY" })
    >>> ed.engine_full
    'MergeTree() PARTITION BY toYear(timestamp) ORDER BY (timestamp, cityHash64(location)) SAMPLE BY cityHash64(location) SETTINGS index_granularity = 32, index_granularity_bytes = 2048 TTL toDate(timestamp) + INTERVAL 1 DAY'
    >>> ed.engine
    'MergeTree'
    >>> ed.to_json()
    {'engine_full': 'MergeTree() PARTITION BY toYear(timestamp) ORDER BY (timestamp, cityHash64(location)) SAMPLE BY cityHash64(location) SETTINGS index_granularity = 32, index_granularity_bytes = 2048 TTL toDate(timestamp) + INTERVAL 1 DAY', 'engine': 'MergeTree', 'partition_key': 'toYear(timestamp)', 'sorting_key': 'timestamp, cityHash64(location)', 'sampling_key': 'cityHash64(location)', 'settings': 'index_granularity = 32, index_granularity_bytes = 2048', 'ttl': 'toDate(timestamp) + INTERVAL 1 DAY'}
    >>> ed.to_datafile()
    'ENGINE "MergeTree"\\nENGINE_PARTITION_KEY "toYear(timestamp)"\\nENGINE_SORTING_KEY "timestamp, cityHash64(location)"\\nENGINE_SAMPLING_KEY "cityHash64(location)"\\nENGINE_SETTINGS "index_granularity = 32, index_granularity_bytes = 2048"\\nENGINE_TTL "toDate(timestamp) + INTERVAL 1 DAY"'

    >>> ed = TableDetails({"engine_full": "Join(ANY, LEFT, id)", "engine": "Join", "partition_key": "", "sorting_key": "", "primary_key": "", "sampling_key": ""})
    >>> ed.engine_full
    'Join(ANY, LEFT, id)'
    >>> ed.engine
    'Join'
    >>> ed.to_json()
    {'engine_full': 'Join(ANY, LEFT, id)', 'engine': 'Join', 'join_strictness': 'ANY', 'join_type': 'LEFT', 'key_columns': 'id'}
    >>> ed.to_datafile()
    'ENGINE "Join"\\nENGINE_JOIN_STRICTNESS "ANY"\\nENGINE_JOIN_TYPE "LEFT"\\nENGINE_KEY_COLUMNS "id"'
    '''
    def __init__(self, details):
        self.details = details or {}

    @property
    def engine_full(self):
        _engine_full = self.details.get('engine_full', None)
        if not _engine_full:
            return None
        _engine_full = _engine_full.replace(' SETTINGS index_granularity = 8192', '')
        return engine_replicated_to_local(_engine_full)

    @property
    def engine(self):
        _engine = self.details.get('engine', None)
        return _engine and _engine.replace('Replicated', '')

    @property
    def version(self):
        _version = self.details.get('version', None)
        return _version

    def is_mergetree_family(self):
        return self.engine and 'mergetree' in self.engine.lower()

    def supports_alter_add_column(self):
        return self.is_mergetree_family() or self.engine.lower() == "null"

    def is_replacing_engine(self):
        if self.engine:
            engine_lower = self.engine.lower()
            is_aggregating = 'aggregatingmergetree' in engine_lower
            is_replacing = 'replacingmergetree' in engine_lower
            is_collapsing = 'collapsingmergetree' in engine_lower
            return is_aggregating or is_replacing or is_collapsing
        return False

    @property
    def partition_key(self):
        return self.details.get('partition_key', None)

    @property
    def sorting_key(self):
        _sorting_key = self.details.get('sorting_key', None)
        if self.is_replacing_engine() and not _sorting_key:
            raise ValueError(f'SORTING_KEY must be defined for the {self.engine} engine')
        if self.is_mergetree_family():
            return _sorting_key or 'tuple()'
        return _sorting_key

    @property
    def primary_key(self):
        _primary_key = self.details.get('primary_key', None)
        if self.sorting_key == _primary_key:
            return None
        return _primary_key

    @property
    def sampling_key(self):
        return self.details.get('sampling_key', None)

    @property
    def settings(self):
        settings = self.details.get('settings', None)
        if settings and settings.strip().lower() != 'index_granularity = 8192':
            return settings

    @property
    def ttl(self):
        return self.details.get('ttl', None)

    @property
    def ver(self):
        _ver = self.details.get('ver', None)
        return _ver

    @property
    def columns(self):
        _columns = self.details.get('columns', None)
        return _columns

    @property
    def sign(self):
        _sign = self.details.get('sign', None)
        return _sign

    @property
    def join_strictness(self):
        _join_strictness = self.details.get('join_strictness', None)
        return _join_strictness

    @property
    def join_type(self):
        _join_type = self.details.get('join_type', None)
        return _join_type

    @property
    def key_columns(self):
        _key_columns = self.details.get('key_columns', None)
        return _key_columns

    @property
    def statistics(self):
        return {
            'bytes': self.details.get('total_bytes', None),
            'row_count': self.details.get('total_rows', None),
        }

    def to_json(self, exclude=None):
        d = {
            'engine_full': self.engine_full,
            'engine': self.engine,
        }
        if self.partition_key:
            d['partition_key'] = self.partition_key
        if self.sorting_key:
            d['sorting_key'] = self.sorting_key
        if self.primary_key:
            d['primary_key'] = self.primary_key
        if self.sampling_key:
            d['sampling_key'] = self.sampling_key
        if self.settings:
            d['settings'] = self.settings
        if self.ttl:
            d['ttl'] = self.ttl.strip()
        if self.engine_full:
            engine_params = engine_params_from_engine_full(self.engine_full)
            d = {**d, **engine_params}

        if exclude:
            for attr in exclude:
                if attr in d:
                    del d[attr]

        return d

    def to_datafile(self):
        d = self.to_json()
        del d['engine_full']
        engine = d['engine']
        del d['engine']
        if engine:
            return '\n'.join([f'ENGINE "{engine}"'] + [f'ENGINE_{k.upper()} "{v}"' for k, v in d.items() if v])
        return ''


def engine_config(name, params=None, options=None):
    params = params or []
    options = options or []
    return (name, (params, options))


class EngineOption:
    def __init__(self, name=None, sql=None, required=None, default_value=None, is_valid=None):
        self.name = name
        self.sql = sql
        self.required = required
        self.default_value = default_value
        self.is_valid = is_valid


class EngineParam:
    def __init__(self, name=None, required=None, default_value=None, is_valid=None):
        self.name = name
        self.required = required
        self.default_value = default_value
        self.is_valid = is_valid


def column_is_valid(columns, column_name):
    schema_columns = [col_name(c['name'], backquotes=False) for c in columns]
    if column_name not in schema_columns:
        raise ValueError(f"'{column_name}' column is not present in schema")
    return col_name(column_name, backquotes=False)


def columns_are_valid(columns, column_names):
    schema_columns = [col_name(c['name'], backquotes=False) for c in columns]
    new_column_names = []
    for column_name in [x.strip() for x in column_names.split(',')]:
        if column_name not in schema_columns:
            raise ValueError(
                f"'{column_name}' column is not present in schema")
        new_column_names.append(col_name(column_name, backquotes=False))
    return ', '.join(new_column_names)


def sorting_key_is_valid(columns, value):
    INVALID_SORTING_KEYS = ['tuple()']

    if not value:
        raise ValueError("Sorting key can not be empty")
    if value in INVALID_SORTING_KEYS:
        raise ValueError(f"'{value}' is not a valid as a sorting key")
    return value


# [PARTITION BY expr]
# [ORDER BY expr]
# [PRIMARY KEY expr]
# [SAMPLE BY expr]
# [TTL expr [DELETE|TO DISK 'xxx'|TO VOLUME 'xxx'], ...]
# [SETTINGS name=value, ...]
MERGETREE_OPTIONS = [
    EngineOption(name='partition_key', sql='PARTITION BY'),
    EngineOption(name='sorting_key', sql='ORDER BY', default_value='tuple()'),
    EngineOption(name='primary_key', sql='PRIMARY KEY'),
    EngineOption(name='sampling_key', sql='SAMPLE BY'),
    EngineOption(name='ttl', sql='TTL'),
    EngineOption(name='settings', sql='SETTINGS'),
]
REPLACINGMERGETREE_OPTIONS = [
    EngineOption(name='partition_key', sql='PARTITION BY'),
    EngineOption(name='sorting_key', sql='ORDER BY', required=True, is_valid=sorting_key_is_valid),
    EngineOption(name='primary_key', sql='PRIMARY KEY'),
    EngineOption(name='sampling_key', sql='SAMPLE BY'),
    EngineOption(name='ttl', sql='TTL'),
    EngineOption(name='settings', sql='SETTINGS'),
]
ENABLED_ENGINES = [
    # MergeTree()
    engine_config('MergeTree', options=MERGETREE_OPTIONS),
    # ReplacingMergeTree([ver])
    engine_config('ReplacingMergeTree', [
        EngineParam(name='ver', is_valid=column_is_valid)
    ], REPLACINGMERGETREE_OPTIONS),
    # SummingMergeTree([columns])
    engine_config('SummingMergeTree', [
        # This should check the columns are numeric ones
        EngineParam(name='columns', is_valid=columns_are_valid),
    ], MERGETREE_OPTIONS),
    # AggregatingMergeTree()
    engine_config('AggregatingMergeTree', options=REPLACINGMERGETREE_OPTIONS),
    # CollapsingMergeTree(sign)
    engine_config('CollapsingMergeTree', [
        EngineParam(name='sign', required=True, is_valid=column_is_valid)
    ], REPLACINGMERGETREE_OPTIONS),
    # VersionedCollapsingMergeTree(sign, version)
    engine_config('VersionedCollapsingMergeTree', [
        EngineParam(name='sign', required=True, is_valid=column_is_valid),
        EngineParam(name='version', required=True, is_valid=column_is_valid),
    ], MERGETREE_OPTIONS),
    # Join(join_strictness, join_type, k1[, k2, ...])
    engine_config('Join', [
        # https://github.com/ClickHouse/ClickHouse/blob/fa8e4e4735b932f08b6beffcb2d069b72de34401/src/Storages/StorageJoin.cpp
        EngineParam(name='join_strictness', required=True,
                    is_valid=['ANY', 'ALL', 'SEMI', 'ANTI']),
        EngineParam(name='join_type', required=True, is_valid=[
                    'LEFT', 'INNER', 'RIGHT', 'FULL']),
        EngineParam(name='key_columns', required=True,
                    is_valid=columns_are_valid),
    ]),
    # Null()
    engine_config('Null'),
]


def get_engine_config(engine):
    for (name, config) in ENABLED_ENGINES:
        if engine.lower() == name.lower():
            return (name, config)
    return None


def engine_params(columns, params, args):
    params_values = []
    for p in params:
        if p.required and p.name not in args:
            raise ValueError(f"Missing required parameter '{p.name}'")
        param_value = args.get(p.name, None) or p.default_value
        if not param_value:
            continue
        if p.is_valid:
            check_is_valid(
                valid_check=p.is_valid,
                check_type='parameter',
                columns=columns,
                name=p.name,
                value=param_value)
        params_values.append(param_value)
    return params_values


def engine_options(columns, options, args):
    options_values = []
    for o in options:
        if o.required and o.name not in args:
            raise ValueError(f"Missing required option '{o.name}'")
        option_value = args.get(o.name) or o.default_value
        if o.is_valid:
            check_is_valid(
                valid_check=o.is_valid,
                check_type='option',
                columns=columns,
                name=o.name,
                value=option_value)
        if option_value:
            if o.sql.lower() == 'settings':
                options_values.append(f"{o.sql} {option_value}")
            else:
                options_values.append(f"{o.sql} ({option_value})")
    return options_values


def check_is_valid(valid_check, check_type, columns, name, value):
    if callable(valid_check):
        try:
            new_value = valid_check(columns, value)
            if new_value:
                value = new_value
        except Exception as e:
            raise ValueError(f"Invalid value '{value}' for {check_type} '{name}', reason: {e}")
    else:
        if value not in valid_check:
            raise ValueError(f"Invalid value '{value}' for {check_type} '{name}', valid values are: {', '.join(valid_check)}")


def build_engine(engine, columns, params, options, args):
    return f"{engine}({', '.join(engine_params(columns, params, args))}) {' '.join(engine_options(columns, options, args))}".strip()


def engine_full_from_dict(engine: str, args: dict, schema: str = None, columns: list = None):
    """
    >>> schema = ''
    >>> engine_full_from_dict('wadus', {}, schema=schema)
    Traceback (most recent call last):
    ...
    ValueError: Engine wadus is not supported, supported engines include: MergeTree, ReplacingMergeTree, SummingMergeTree, AggregatingMergeTree, CollapsingMergeTree, VersionedCollapsingMergeTree, Join, Null

    >>> schema = ''
    >>> engine_full_from_dict('null', {}, schema=schema)
    'Null()'
    >>> schema = ''
    >>> engine_full_from_dict('null', {}, columns=[])
    'Null()'

    >>> schema = 'cid Int32'
    >>> engine_full_from_dict('Join', {'join_strictness': 'ANY', 'join_type': 'LEFT', 'key_columns': 'cid'}, schema=schema)
    'Join(ANY, LEFT, cid)'
    >>> engine_full_from_dict('Join', {'join_strictness': 'ANY', 'join_type': 'LEFT', 'key_columns': 'cid'}, columns=[{'name': 'cid', 'type': 'Int32', 'codec': None, 'default_value': None, 'nullable': False, 'normalized_name': 'cid'}])
    'Join(ANY, LEFT, cid)'
    >>> schema = 'cid1 Int32, cid2 Int8'
    >>> engine_full_from_dict('Join', {'join_strictness': 'ANY', 'join_type': 'LEFT', 'key_columns': 'cid1, cid2'}, schema=schema)
    'Join(ANY, LEFT, cid1, cid2)'
    >>> engine_full_from_dict('Join', {'join_strictness': 'ANY', 'join_type': 'OUTER', 'key_columns': 'cid'}, schema=schema)
    Traceback (most recent call last):
    ...
    ValueError: Invalid value 'OUTER' for parameter 'join_type', valid values are: LEFT, INNER, RIGHT, FULL

    >>> schema = ''
    >>> engine_full_from_dict('MergeTree', {}, schema=schema)
    'MergeTree() ORDER BY (tuple())'
    >>> engine_full_from_dict('MergeTree', {'sorting_key': 'local_date, cod_store'}, schema=schema)
    'MergeTree() ORDER BY (local_date, cod_store)'
    >>> engine_full_from_dict('MergeTree', {'partition_key': 'toDate(timestamp)', 'sorting_key': 'local_date, cod_store', 'settings': 'index_granularity = 32, index_granularity_bytes = 2048', 'ttl': 'toDate(local_date) + INTERVAL 1 DAY'}, schema=schema)
    'MergeTree() PARTITION BY (toDate(timestamp)) ORDER BY (local_date, cod_store) TTL (toDate(local_date) + INTERVAL 1 DAY) SETTINGS index_granularity = 32, index_granularity_bytes = 2048'

    >>> schema = ''
    >>> engine_full_from_dict('CollapsingMergeTree', {'sign': 'sign_column'}, schema=schema)
    Traceback (most recent call last):
    ...
    ValueError: Invalid value 'sign_column' for parameter 'sign', reason: 'sign_column' column is not present in schema

    >>> schema = 'sign_column Int8'
    >>> engine_full_from_dict('CollapsingMergeTree', {'sign': 'sign_column'}, schema=schema)
    Traceback (most recent call last):
    ...
    ValueError: Missing required option 'sorting_key'

    >>> schema = 'sign_column Int8, key_column Int8'
    >>> engine_full_from_dict('CollapsingMergeTree', {'sign': 'sign_column', 'sorting_key': 'key_column'}, schema=schema)
    'CollapsingMergeTree(sign_column) ORDER BY (key_column)'

    >>> columns=[]
    >>> columns.append({'name': 'sign_column', 'type': 'Int8', 'codec': None, 'default_value': None, 'nullable': False, 'normalized_name': 'sign_column'})
    >>> columns.append({'name': 'key_column', 'type': 'Int8', 'codec': None, 'default_value': None, 'nullable': False, 'normalized_name': 'key_column'})
    >>> engine_full_from_dict('CollapsingMergeTree', {'sign': 'sign_column', 'sorting_key': 'key_column' }, columns=columns)
    'CollapsingMergeTree(sign_column) ORDER BY (key_column)'

    >>> schema = 'sign_column Int8'
    >>> engine_full_from_dict('AggregatingMergeTree', {}, schema=schema)
    Traceback (most recent call last):
    ...
    ValueError: Missing required option 'sorting_key'

    >>> columns=[]
    >>> columns.append({'name': 'key_column', 'type': 'Int8', 'codec': None, 'default_value': None, 'nullable': False, 'normalized_name': 'key_column'})
    >>> engine_full_from_dict('AggregatingMergeTree', { 'sorting_key': 'key_column' }, columns=columns)
    'AggregatingMergeTree() ORDER BY (key_column)'

    >>> schema = 'ver_column Int8, key_column Int8'
    >>> engine_full_from_dict('ReplacingMergeTree', {}, schema=schema)
    Traceback (most recent call last):
    ...
    ValueError: Missing required option 'sorting_key'

    >>> engine_full_from_dict('ReplacingMergeTree', {'sorting_key': 'key_column'}, schema=schema)
    'ReplacingMergeTree() ORDER BY (key_column)'

    >>> engine_full_from_dict('ReplacingMergeTree', {'ver': 'ver_column', 'sorting_key': 'key_column'}, schema=schema)
    'ReplacingMergeTree(ver_column) ORDER BY (key_column)'

    >>> engine_full_from_dict('ReplacingMergeTree', {'ver': 'other_column'}, schema=schema)
    Traceback (most recent call last):
    ...
    ValueError: Invalid value 'other_column' for parameter 'ver', reason: 'other_column' column is not present in schema

    >>> schema = 'col0 Int8, col1 Int8, col2 Int8'
    >>> engine_full_from_dict('SummingMergeTree', {}, schema=schema)
    'SummingMergeTree() ORDER BY (tuple())'
    >>> engine_full_from_dict('SummingMergeTree', {'columns': 'col0'}, schema=schema)
    'SummingMergeTree(col0) ORDER BY (tuple())'
    >>> engine_full_from_dict('SummingMergeTree', {'columns': 'col0, col2'}, schema=schema)
    'SummingMergeTree(col0, col2) ORDER BY (tuple())'
    >>> engine_full_from_dict('SummingMergeTree', {'columns': 'col1, other_column'}, schema=schema)
    Traceback (most recent call last):
    ...
    ValueError: Invalid value 'col1, other_column' for parameter 'columns', reason: 'other_column' column is not present in schema
    >>> engine_full_from_dict('SummingMergeTree', {'columns': 'col1, other_column'}, schema=schema, columns=[])
    Traceback (most recent call last):
    ...
    ValueError: You can not use 'schema' and 'columns' at the same time
    """
    if schema is not None and columns is not None:
        raise ValueError("You can not use 'schema' and 'columns' at the same time")
    engine_config = get_engine_config(engine)

    if not engine_config:
        raise ValueError(
            f"Engine {engine} is not supported, supported engines include: {', '.join([e[0] for e in ENABLED_ENGINES])}")
    name, (params, options) = engine_config
    if columns is None:
        columns = parse_table_structure(schema)

    for arg in args:
        if not hasattr(TableDetails, arg):
            raise ValueError(f"engine_{arg} is not a valid option")

    return build_engine(name, columns, params, options, args)


def engine_params_from_engine_full(engine_full: str):
    """
    >>> engine_params_from_engine_full("Null()")
    {}
    >>> engine_params_from_engine_full("Join(ANY, LEFT, id)")
    {'join_strictness': 'ANY', 'join_type': 'LEFT', 'key_columns': 'id'}
    >>> engine_params_from_engine_full("Join(ANY, LEFT, k1, k2)")
    {'join_strictness': 'ANY', 'join_type': 'LEFT', 'key_columns': 'k1, k2'}
    >>> engine_params_from_engine_full("AggregatingMergeTree('/clickhouse/tables/{layer}-{shard}/d_f837aa.sales_by_country_rt__v0_staging_t_00c3091e7530472caebda05e97288a1d', '{replica}') PARTITION BY toYYYYMM(date) ORDER BY (purchase_location, cod_device, date) SETTINGS index_granularity = 8192")
    {}
    >>> engine_params_from_engine_full("ReplicatedSummingMergeTree('/clickhouse/tables/{layer}-{shard}/d_abcf3e.t_69f9da31f4524995b8911e1b24c80ab4', '{replica}') PARTITION BY toYYYYMM(date) ORDER BY (date, purchase_location, sku_rank_lc) SETTINGS index_granularity = 8192")
    {}
    >>> engine_params_from_engine_full("ReplicatedSummingMergeTree('/clickhouse/tables/{layer}-{shard}/d_abcf3e.t_69f9da31f4524995b8911e1b24c80ab4', '{replica}', c1, c2) PARTITION BY toYYYYMM(date) ORDER BY (date, purchase_location, sku_rank_lc) SETTINGS index_granularity = 8192")
    {'columns': 'c1, c2'}
    >>> engine_params_from_engine_full("ReplacingMergeTree(insert_date) ORDER BY date")
    {'ver': 'insert_date'}
    >>> engine_params_from_engine_full("ReplicatedReplacingMergeTree('/clickhouse/tables/{layer}-{shard}/d_f837aa.t_d3aaad001dee4d9e9e3067ccb789fb59_n1', '{replica}', insert_date) ORDER BY pk TTL toDate(local_timeplaced) + toIntervalDay(3) SETTINGS index_granularity = 8192")
    {'ver': 'insert_date'}
    >>> engine_params_from_engine_full("ReplicatedVersionedCollapsingMergeTree('/clickhouse/tables/{layer}-{shard}/test.foo', '{replica}', sign_c,version_c) ORDER BY pk TTL toDate(local_timeplaced) + toIntervalDay(3) SETTINGS index_granularity = 8192")
    {'sign': 'sign_c', 'version': 'version_c'}
    """
    engine_full = engine_replicated_to_local(engine_full)
    for engine, (params, _options) in ENABLED_ENGINES:
        if engine_full.startswith(engine):
            m = re.search(rf'{engine}\(([^\)]*)\).*', engine_full)
            params_used = []
            if m:
                params_used = [x.strip() for x in m.group(1).split(',')]
            params_dict = defaultdict(list)
            param = None
            for i, v in enumerate(params_used):
                if i < len(params):
                    param = params[i]
                if param and v:
                    params_dict[param.name].append(v)

            return {k: ', '.join(v) for k, v in params_dict.items()}
    return {}



from tornado.simple_httpclient import SimpleAsyncHTTPClient
from tornado.httpclient import HTTPClientError
import json

PYPY_URL = 'https://pypi.org/pypi/tinybird-cli/json'


class CheckPypi:
    client = SimpleAsyncHTTPClient()

    async def get_latest_version(self):
        try:
            response = await self.client.fetch(PYPY_URL)
            result = json.loads(response.body)
            return result['info']['version']
        except HTTPClientError as e:
            response = e.response
        except Exception as e:
            print(f'Error getting package info: {e}')



import logging
import builtins
from tornado.template import Template
from tornado import escape
from io import StringIO
import ast
import linecache
from tornado.util import ObjectDict, exec_in, unicode_type
from .datatypes import testers
from datetime import datetime


class SQLTemplateCustomError(Exception):
    def __init__(self, err, code=400):
        self.code = code
        self.err = err
        super().__init__(err)

# t = Template(""" SELECT * from test where lon between {{Float32(lon1, 0)}} and {{Float32(lon2, 0)}} """)
# names = get_var_names(t)
# print(generate(t, **{x: '' for x in names}))

# t = Template(""" SELECT * from test where lon between {{lon1}} and {{lon2}} """)
# names = get_var_names(t)
# replace_vars_smart(t)
# print(generate(t, **{x: '' for x in names}))


DEFAULT_PARAM_NAMES = ['format', 'q']

parameter_types = [
    'String',
    'Boolean',
    'DateTime',
    'Date',
    'Float32',
    'Float64',
    'Int8',
    'Int16',
    'Int32',
    'Int64',
    'UInt8',
    'UInt16',
    'UInt32',
    'UInt64',
    'Array'
]

parameter_openapi_types = {
    'String': 'string',
    'Boolean': 'boolean',
    'DateTime': 'string',
    'Date': 'string',
    'Float32': 'number',
    'Float64': 'number',
    'Int8': 'integer',
    'Int16': 'integer',
    'Int32': 'integer',
    'Int64': 'integer',
    'UInt8': 'integer',
    'UInt16': 'integer',
    'UInt32': 'integer',
    'UInt64': 'integer',
    'Array': 'array'
}


def transform_type(tester, transform, placeholder=None, required=None, description=None, enum=None, example=None, format=None):
    def _f(x, default=None, defined=True, required=None, description=None, enum=None, example=None, format=None):
        if type(x) == Placeholder:
            if default:
                x = default
            else:
                x = placeholder
        elif x is None:
            x = default
            if x is None:
                if defined:
                    raise ValueError("invalid parameter")
                else:
                    return None
        if tester == 'String':
            if x is not None:
                return transform(x)
        elif testers[tester](str(x)):
            return transform(x)
        raise ValueError(f"Error validating '{x}' to type {tester}")
    return _f


def _and(*args, **kwargs):
    operands = {'in': 'in', 'not_in': 'not in', 'gt': '>', 'lt': '<'}

    def _name(k):
        tk = k.rsplit('__', 1)
        return tk[0]

    def _op(k):
        tk = k.rsplit('__', 1)
        if len(tk) == 1:
            return '='
        else:
            if tk[1] in operands:
                return operands[tk[1]]
            raise ValueError(f"operand {tk[1]} not supported")
    return Expression(' and '.join(
        [f"{_name(k)} {_op(k)} {expression_wrapper(v, k)}" for k, v in kwargs.items() if v is not None]))


def error(s, code=400):
    raise ValueError(s)


def custom_error(s, code=400):
    raise SQLTemplateCustomError(s, code)


class Expression(str):
    pass


class Comment:
    def __init__(self, s):
        self.text = s

    def __str__(self):
        return self.text


class Placeholder:
    def __init__(self, name=None, line=None):
        self.name = name if name else '__placeholder__'
        self.line = line or 'unknown'

    def __str__(self):
        return '__placeholder__'

    def __getitem__(self, i):
        if i > 2:
            raise IndexError()
        return Placeholder()

    def __add__(self, s):
        return Placeholder()

    def __call__(self, *args, **kwargs):
        raise ValueError(f"'{self.name}' is not a valid function, line {self.line}")

    def split(self, ch):
        return [Placeholder(), Placeholder()]

    def startswith(self, c):
        return False


class Symbol:
    def __init__(self, x):
        self.x = x

    def __str__(self):
        return self.x


def columns(x, default=None, fn=None):
    if x is None or type(x) == Placeholder:
        if default is None:
            raise ValueError("Missing columns() default value, use `columns(column_names, 'default_column_name')`")
        x = default

    _columns = [c.strip() for c in x.split(',')]

    if fn:
        return Expression(','.join(f'{fn}({str(column(c, c))}) as {c}' for c in _columns))
    else:
        return Expression(','.join(str(column(c, c)) for c in _columns))


def column(x, default=None):
    if x is None or type(x) == Placeholder:
        if default is None:
            raise ValueError("Missing column() default value, use `column(column_name, 'default_column_name')`")
        x = default
    return Symbol("`" + sqlescape(x) + "`")


def symbol(x, quote='`'):
    if type(x) == Placeholder:
        return Symbol("`placeholder`")
    return Symbol(quote + sqlescape(x) + quote)


def boolean(x, default=None):
    if x is None:
        raise ValueError('wrong value')
    if type(x) == Placeholder:
        if default is not None:
            return boolean(default)
        return True

    elif type(x) == bool:
        return x
    elif x.lower() == 'true':
        return True
    elif x.lower() == 'false':
        return False
    if testers['Integer'](x):
        return int(x)

    if '' == x:
        return False
    return True


def defined(x=None):
    if type(x) == Placeholder or x is None:
        return False
    return True


def array_type(types):  # noqa: C901
    def _f(x, _type=None, default=None, defined=True, required=None, description=None, enum=None, example=None, format=None):
        if type(x) == Placeholder:
            if default:
                x = default
            else:
                if _type and _type in types:
                    x = ','.join(map(str, [types[_type](x) for _ in range(2)]))
                else:
                    x = ','.join([f'__placeholder__{i}' for i in range(2)])
        elif x is None:
            x = default
            if x is None:
                if defined:
                    raise ValueError("invalid parameter")
                else:
                    return None
        values = []
        for i, t in enumerate(x.split(',')):
            if _type in testers:
                if testers[_type](str(t)):
                    values.append(expression_wrapper(
                        types[_type](t), f'{x}[{i}]'))
                else:
                    raise ValueError(
                        f"Error validating {x}[{i}]({t}) to type {_type}")
            else:
                values.append(expression_wrapper(
                    types.get(_type, lambda x: x)(t), f'{x}[{i}]'))
        return Expression(f"({','.join(map(str, values))})")
    return _f


def sql_unescape(x, what=''):
    """
    unescapes specific characters in a string. It allows to allow some
    special characters to be used, for example in like condictionals

    {{sql_unescape(String(like_filter), '%')}}


    >>> sql_unescape('testing%', '%')
    "'testing%'"
    >>> sql_unescape('testing%', '$')
    "'testing\\\\%'"
    """
    return Expression("'" + sqlescape(x).replace(f'\\{what}', what) + "'")


def split_to_array(x, default=''):
    if type(x) == Placeholder or x is None:
        x = default
    return [s.strip() for s in x.split(',')]


def enumerate_with_last(arr):
    """
    >>> enumerate_with_last([1, 2])
    [(False, 1), (True, 2)]
    >>> enumerate_with_last([1])
    [(True, 1)]
    """
    arr_len = len(arr)
    return [(arr_len == i + 1, x) for i, x in enumerate(arr)]


def string_type(x, default=None):
    if type(x) == Placeholder:
        if default:
            x = default
        else:
            x = '__placeholder__'
    return x


def day_diff(d0, d1, default=None):
    """
    >>> day_diff('2019-01-01', '2019-01-01')
    0
    >>> day_diff('2019-01-01', '2019-01-02')
    1
    >>> day_diff('2019-01-02', '2019-01-01')
    1
    >>> day_diff('2019-01-02', '2019-02-01')
    30
    >>> day_diff('2019-02-01', '2019-01-02')
    30
    >>> day_diff(Placeholder(), Placeholder())
    0
    >>> day_diff(Placeholder(), '')
    0
    """
    if type(d0) == Placeholder or type(d1) == Placeholder:
        if default:
            return default
        return 0
    isoformat = "%Y-%m-%d"
    try:
        return abs((datetime.strptime(d1, isoformat) - datetime.strptime(d0, isoformat)).days)
    except Exception:
        raise Exception(
            'invalid date format, it must be ISO format date, e.g. 2018-09-07')


function_list = {
    'columns': columns,
    'table': symbol,
    'TABLE': symbol,
    'error': error,
    'custom_error': custom_error,
    'sql_and': _and,
    'defined': defined,
    'column': column,
    'enumerate_with_last': enumerate_with_last,
    'split_to_array': split_to_array,
    'day_diff': day_diff,
    'sql_unescape': sql_unescape
    # 'enumerate': enumerate
}


def get_transform_types(placeholders=None):
    if placeholders is None:
        placeholders = {}
    types = {
        'bool': boolean,
        'Boolean': boolean,
        'DateTime': transform_type('DateTime', str, placeholders.get('DateTime', None), required=None, description=None, enum=None, example=None, format=None),
        'Date': transform_type('Date', str, placeholders.get('Date', None), required=None, description=None, enum=None, example=None, format=None),
        'Float32': transform_type('Float32', float, placeholders.get('Float32', None), required=None, description=None, enum=None, example=None, format=None),
        'Float64': transform_type('Float64', float, placeholders.get('Float64', None), required=None, description=None, enum=None, example=None, format=None),
        'Int': transform_type('Int32', int, placeholders.get('Int', None), required=None, description=None, enum=None, example=None, format=None),
        'Integer': transform_type('Int32', int, placeholders.get('Int32', None), required=None, description=None, enum=None, example=None, format=None),
        'Int8': transform_type('Int8', int, placeholders.get('Int8', None), required=None, description=None, enum=None, example=None, format=None),
        'Int16': transform_type('Int16', int, placeholders.get('Int16', None), required=None, description=None, enum=None, example=None, format=None),
        'UInt8': transform_type('UInt8', int, placeholders.get('UInt8', None), required=None, description=None, enum=None, example=None, format=None),
        'UInt16': transform_type('UInt16', int, placeholders.get('UInt16', None), required=None, description=None, enum=None, example=None, format=None),
        'UInt32': transform_type('UInt32', int, placeholders.get('UInt32', None), required=None, description=None, enum=None, example=None, format=None),
        'Int32': transform_type('Int32', int, placeholders.get('Int32', None), required=None, description=None, enum=None, example=None, format=None),
        'Int64': transform_type('Int64', int, placeholders.get('Int64', None), required=None, description=None, enum=None, example=None, format=None),
        'UInt64': transform_type('UInt64', int, placeholders.get('UInt64', None), required=None, description=None, enum=None, example=None, format=None),
        'Symbol': symbol,
        'Column': symbol,
        'String': transform_type('String', str, placeholder='__placeholder__', required=None, description=None, enum=None, example=None, format=None)
    }
    types['Array'] = array_type(types)
    types.update(function_list)
    return types


type_fns = get_transform_types()
type_fns_check = get_transform_types({
    'DateTime': '2019-01-01 00:00:00',
    'Date': '2019-01-01',
    'Float32': 0.0,
    'Float64': 0.0,
    'Int': 0,
    'Int8': 0,
    'Int16': 0,
    'UInt8': 0,
    'UInt16': 0,
    'UInt32': 0,
    'Int32': 0,
    'Int64': 0,
    'UInt64': 0,
    'Symbol': 'symbol'
})


# from https://github.com/elouajib/sqlescapy/
# MIT license
def sqlescape(str):
    return str.translate(
        str.maketrans({
            "\0": "\\0",
            "\r": "\\r",
            "\x08": "\\b",
            "\x09": "\\t",
            "\x1a": "\\z",
            "\n": "\\n",
            "\r": "\\r",
            "\"": "",
            "'": "\\'",
            "\\": "\\\\",
            "%": "\\%",
            "`": "\\`"
        }))


def expression_wrapper(x, name):
    if type(x) in (unicode_type, bytes, str):
        return "'" + sqlescape(x) + "'"
    elif type(x) == Placeholder:
        return "'__placeholder__'"
    elif type(x) == Comment:
        return "-- {x} \n"
    if x is None:
        raise ValueError(f'expression "{name}" evaluated to null')
    return x


_namespace = {
    "column": column,
    "symbol": symbol,
    "error": error,
    "custom_error": custom_error,
    "_tt_utf8": escape.utf8,  # for internal use
    "_tt_string_types": (unicode_type, bytes),
    "xhtml_escape": lambda x: x,
    "expression_wrapper": expression_wrapper,
    # disable __buildins__ and some other functions
    # they raise a pretty non undestandable error but if someone
    # is using them they know what they are trying to do
    # read https://anee.me/escaping-python-jails-849c65cf306e on how to escape from python jails
    "__buildins__": {},
    "__import__": {},
    '__debug__': {},
    '__doc__': {},
    '__name__': {},
    '__package__': {},
    "open": {},
    "close": {},
    "print": {}
}


reserved_vars = set(['_tt_tmp', '_tt_append', 'isinstance', 'str', 'error', 'custom_error'] + list(vars(builtins)))
for p in DEFAULT_PARAM_NAMES:  # we handle these in an specific manner
    reserved_vars.discard(p)  # `format` is part of builtins
error_vars = ['error', 'custom_error']


def generate(self, template_execution_results=None, **kwargs):
    """Generate this template with the given arguments."""
    if template_execution_results is None:
        template_execution_results = {}
    namespace = {}

    def set_max_threads(x):
        try:
            template_execution_results['max_threads'] = int(x)
            return Expression(f'-- max_threads {x}\n')
        except Exception:
            return Expression(f'-- max_threads: wrong argument {x}\n')

    def set_backend_hint(hint):
        template_execution_results['backend_hint'] = hint
        return Expression(f'-- backend_hint {hint}\n')

    namespace.update(_namespace)
    namespace.update(kwargs)
    namespace.update({
        # __name__ and __loader__ allow the traceback mechanism to find
        # the generated source code.
        "__name__": self.name.replace('.', '_'),
        "__loader__": ObjectDict(get_source=lambda name: self.code),
        "max_threads": set_max_threads,
        "backend_hint": set_backend_hint,
    })

    exec_in(self.compiled, namespace)
    execute = namespace["_tt_execute"]
    # Clear the traceback module's cache of source data now that
    # we've generated a new template (mainly for this module's
    # unittests, where different tests reuse the same name).
    linecache.clearcache()

    try:
        return execute()
    except SQLTemplateCustomError as e:
        raise e
    except Exception as e:
        if 'x' in namespace and namespace['x'] and namespace['x'].line:
            line = namespace['x'].line
            raise ValueError(f'{e}, line {line}')
        raise e


class CodeWriter(object):
    def __init__(self, file, template):
        self.file = file
        self.current_template = template
        self.apply_counter = 0
        self._indent = 0

    def indent_size(self):
        return self._indent

    def indent(self):
        class Indenter(object):
            def __enter__(_):
                self._indent += 1
                return self

            def __exit__(_, *args):
                assert self._indent > 0
                self._indent -= 1

        return Indenter()

    def write_line(self, line, line_number, indent=None):
        if indent is None:
            indent = self._indent
        line_comment = '  # %s:%d' % ('<generated>', line_number)
        print("    " * indent + line + line_comment, file=self.file)


def get_var_names(t):
    def _n(chunks, v):
        for x in chunks:
            line_number = x.line
            if type(x).__name__ == '_ChunkList':
                _n(x.chunks, v)
            elif type(x).__name__ == "_Expression":
                c = compile(x.expression, "<string>", "exec", dont_inherit=True)
                variable_names = [x for x in c.co_names if x not in _namespace and x not in reserved_vars]
                v += list(map(lambda variable: {'line': line_number, 'name': variable}, variable_names))
            elif type(x).__name__ == '_ControlBlock':
                from io import StringIO
                buffer = StringIO()
                writer = CodeWriter(buffer, t)
                x.generate(writer)
                c = compile(buffer.getvalue(), "<string>", "exec", dont_inherit=True)
                variable_names = [x for x in c.co_names if x not in _namespace and x not in reserved_vars]
                v += list(map(lambda variable: {'line': line_number, 'name': variable}, variable_names))
                _n(x.body.chunks, v)

    var = []
    _n(t.file.body.chunks, var)
    return var


def get_var_data(content):  # noqa: C901
    def node_to_value(x):
        if type(x) in (ast.Bytes, ast.Str):
            return x.s
        elif type(x) == ast.Num:
            return x.n
        elif type(x) == ast.NameConstant:
            return x.value
        elif type(x) == ast.Name:
            return x.id
        elif type(x) == ast.List:
            return [elem.s for elem in x.elts]
        elif type(x) == ast.BinOp:
            # in this case there could be several variables
            # if that's the case the left one is the main
            r = node_to_value(x.left)
            if not r:
                r = node_to_value(x.right)
            return r
        elif type(x) == ast.Constant:
            return x.value
        else:
            try:
                return x.id
            except Exception:
                # don't let this ruin the parsing
                pass
        return None

    def _w(parsed):
        vars = {}
        for node in ast.walk(parsed):
            if isinstance(node, ast.Call) and isinstance(node.func, ast.Name):
                try:
                    func = node.func.id
                    # parse function args
                    args = []
                    for x in node.args:
                        if type(x) == ast.Call:
                            vars.update(_w(x))
                        else:
                            args.append(node_to_value(x))

                    kwargs = {}
                    for x in node.keywords:
                        kwargs[x.arg] = node_to_value(x.value)

                    if func == 'defined':
                        # define it as string and if it's used later it'll become something else
                        if len(args) > 0 and args[0] not in vars:
                            vars[args[0]] = {"type": 'String', "default": None}
                    elif func == 'Array':
                        if 'default' not in kwargs:
                            default = kwargs.get('default', args[2] if len(args) > 2 and args[2] else None)
                            kwargs['default'] = default
                        vars[args[0]] = {
                            "type": f'Array({args[1]})' if len(args) > 1 else 'Array(String)',
                            **kwargs
                        }
                    elif func in parameter_types:
                        # avoid variable names to be None
                        if args[0] is not None:
                            # if this is a cast use the function name to get the type
                            if 'default' not in kwargs:
                                default = kwargs.get('default', args[1] if len(args) > 1 else None)
                                kwargs['default'] = default
                            vars[args[0]] = {"type": func, **kwargs}
                except Exception as e:
                    # if we find a problem parsing, let the parsing continue
                    logging.error(f"pipe parsing problem {content}:  {e}")
            elif isinstance(node, ast.Name):
                # when parent node is a call it means it's managed by the Call workflow (see above)
                is_cast = isinstance(node.parent, ast.Call) and isinstance(node.parent.func, ast.Name) and node.parent.func.id in parameter_types
                is_reserved_name = node.id in reserved_vars or node.id in function_list or node.id in _namespace
                if (not isinstance(node.parent, ast.Call) or not is_cast) and not is_reserved_name:
                    vars[node.id] = {"type": 'String', 'default': None}

        return vars

    parsed = ast.parse(content)
    # calculate parents for each node for later checks
    for node in ast.walk(parsed):
        for child in ast.iter_child_nodes(node):
            child.parent = node
    vars = _w(parsed)

    return [dict(name=k, **v) for k, v in vars.items()]


def get_var_names_and_types(t):  # noqa: C901
    """
    >>> get_var_names_and_types(Template("SELECT * FROM filter_value WHERE description = {{String(d, 'test_1')}} AND value = {{Int8(v, 3)}}"))
    [{'name': 'd', 'type': 'String', 'default': 'test_1'}, {'name': 'v', 'type': 'Int8', 'default': 3}]
    >>> get_var_names_and_types(Template("select * from test {% if defined(testing) and defined(testing2) %} where 1 {%end %}"))
    [{'name': 'testing', 'type': 'String', 'default': None}, {'name': 'testing2', 'type': 'String', 'default': None}]
    >>> get_var_names_and_types(Template("select {{Array(cod_stock_source_type,'Int16', defined=False)}}"))
    [{'name': 'cod_stock_source_type', 'type': 'Array(Int16)', 'defined': False, 'default': None}]
    >>> get_var_names_and_types(Template("select {{Array(cod_stock_source_type, defined=False)}}"))
    [{'name': 'cod_stock_source_type', 'type': 'Array(String)', 'defined': False, 'default': None}]
    >>> get_var_names_and_types(Template("select {{cod_stock_source_type}}"))
    [{'name': 'cod_stock_source_type', 'type': 'String', 'default': None}]
    >>> get_var_names_and_types(Template("select {{String(cod_stock_source_type, 'test')}}"))
    [{'name': 'cod_stock_source_type', 'type': 'String', 'default': 'test'}]
    >>> get_var_names_and_types(Template("select {{split_to_array(test)}}"))
    [{'name': 'test', 'type': 'String', 'default': None}]
    >>> get_var_names_and_types(Template("select {{'pepe'.startswith('p')}}"))
    []
    >>> get_var_names_and_types(Template("select {{String(test + 'abcd', 'default_value')}}"))
    [{'name': 'test', 'type': 'String', 'default': None}]
    >>> get_var_names_and_types(Template("SELECT * FROM filter_value WHERE description = {{String(d, 'test_1', description='test', required=True)}} AND value = {{Int8(v, 3, format='number', example='1')}}"))
    [{'name': 'd', 'type': 'String', 'description': 'test', 'required': True, 'default': 'test_1'}, {'name': 'v', 'type': 'Int8', 'format': 'number', 'example': '1', 'default': 3}]
    >>> get_var_names_and_types(Template("SELECT * FROM filter_value WHERE description = {{String(d, default='test_1', description='test')}}"))
    [{'name': 'd', 'type': 'String', 'default': 'test_1', 'description': 'test'}]
    >>> get_var_names_and_types(Template("select {{Array(cod_stock_source_type, 'Int16', default='1', defined=False)}}"))
    [{'name': 'cod_stock_source_type', 'type': 'Array(Int16)', 'default': '1', 'defined': False}]
    >>> get_var_names_and_types(Template('select {{symbol(split_to_array(attr, "amount_net")[0] + "_intermediate" )}}'))
    [{'name': 'attr', 'type': 'String', 'default': None}]
    >>> get_var_names_and_types(Template("SELECT * FROM filter_value WHERE description = {{Float32(with_value, 0.1)}} AND description = {{Float32(zero, 0)}} AND value = {{Float32(no_default)}}"))
    [{'name': 'with_value', 'type': 'Float32', 'default': 0.1}, {'name': 'zero', 'type': 'Float32', 'default': 0}, {'name': 'no_default', 'type': 'Float32', 'default': None}]
    >>> get_var_names_and_types(Template('''SELECT * FROM abcd WHERE hotel_id <> 0 {% if defined(date_from) %} AND script_created_at > {{DateTime(date_from, '2020-09-09 10:10:10', description="This is a description", required=True)(date_from, '2020-09-09', description="Filter script alert creation date", required=False)}} {% end %}'''))
    [{'name': 'date_from', 'type': 'String', 'default': None}, {'name': 'date_from', 'type': 'String', 'default': None}]

    """

    def _n(chunks, v):
        for x in chunks:
            if type(x).__name__ == '_ChunkList':
                _n(x.chunks, v)
            elif type(x).__name__ == "_Expression":
                var_data = get_var_data(x.expression)
                if var_data:
                    v += var_data
            elif type(x).__name__ == '_ControlBlock':
                buffer = StringIO()
                writer = CodeWriter(buffer, t)
                x.generate(writer)
                var_data = get_var_data(buffer.getvalue())
                if var_data:
                    v += var_data
                _n(x.body.chunks, v)
    var = []
    _n(t.file.body.chunks, var)
    return var


def wrap_vars(t):
    def _n(chunks, v):
        for x in chunks:
            if type(x).__name__ == '_ChunkList':
                _n(x.chunks, v)
            elif type(x).__name__ == "_Expression":
                x.expression = 'expression_wrapper(' + x.expression + \
                    ',"""' + x.expression.replace('"', '\\"') + '""")'
            elif type(x).__name__ == '_ControlBlock':
                _n(x.body.chunks, v)

    var = []
    _n(t.file.body.chunks, var)
    t.code = t._generate_python(t.loader)
    try:
        t.compiled = compile(
            escape.to_unicode(t.code),
            "%s.generated.py" % t.name.replace('.', '_'),
            "exec", dont_inherit=True)
    except Exception:
        # formatted_code = _format_code(t.code).rstrip()
        # app_log.error("%s code:\n%s", t.name, formatted_code)
        raise

    return var


def get_used_tables_in_template(sql):
    """
    >>> get_used_tables_in_template("select * from {{table('test')}}")
    ['test']
    >>> get_used_tables_in_template("select * from {%if x %}{{table('test')}}{%else%}{{table('test2')}}{%end%}")
    ['test', 'test2']
    """
    t = Template(sql)

    def _n(chunks, tables):
        for x in chunks:
            if type(x).__name__ == "_Expression":
                c = compile(x.expression, "<string>",
                            "exec", dont_inherit=True)
                v = [
                    x.lower() for x in c.co_names if x not in _namespace and x not in reserved_vars]
                if 'table' in v:
                    def _t(*args, **kwargs):
                        return str(args[0])
                    n = {'table': _t, 'TABLE': _t}
                    e = "_tt_tmp = %s" % x.expression
                    exec_in(e, n)
                    tables += [n['_tt_tmp']]
            elif type(x).__name__ == '_ControlBlock':
                _n(x.body.chunks, tables)

    tables = []
    _n(t.file.body.chunks, tables)
    return tables


def render_sql_template(sql, variables=None, test_mode=False, template_execution_results=None):
    """
    >>> render_sql_template("select * from table where str = {{foo}}", { 'foo': 'test' })
    "select * from table where str = 'test'"
    >>> render_sql_template("select * from table where f = {{foo}}", { 'foo': 1.0 })
    'select * from table where f = 1.0'
    >>> render_sql_template("select * from table where f = {{Float32(foo)}}", { 'foo': 1 })
    'select * from table where f = 1.0'
    >>> render_sql_template("select * from table where f = {{foo}}", { 'foo': "';drop table users;" })
    "select * from table where f = '\\\\';drop table users;'"
    >>> render_sql_template("select * from {{symbol(foo)}}", { 'foo': 'table-name' })
    'select * from `table-name`'
    >>> render_sql_template("select * from {{Int32(foo)}}", { 'foo': 'non_int' })
    Traceback (most recent call last):
    ...
    ValueError: Error validating 'non_int' to type Int32
    >>> render_sql_template("select * from table where f = {{Float32(foo)}}", test_mode=True)
    'select * from table where f = 0.0'
    >>> render_sql_template("SELECT * FROM query_log__dev where a = {{test}}", test_mode=True)
    "SELECT * FROM query_log__dev where a = '__placeholder__'"
    >>> render_sql_template("SELECT {{test}}", {'token':'testing'})
    Traceback (most recent call last):
    ...
    ValueError: expression "test" evaluated to null
    >>> render_sql_template('{% if test %}SELECT 1{% else %} select 2 {% end %}')
    ' select 2 '
    >>> render_sql_template('{% if Int32(test, 1) %}SELECT 1{% else %} select 2 {% end %}')
    'SELECT 1'
    >>> render_sql_template('{% for v in test %}SELECT {{v}} {% end %}',test_mode=True)
    "SELECT '__placeholder__' SELECT '__placeholder__' SELECT '__placeholder__' "
    >>> render_sql_template("select {{Int32(foo, 1)}}", test_mode=True)
    'select 1'
    >>> render_sql_template("SELECT count() c FROM test_table where a > {{Float32(myvar)}} {% if defined(my_condition) %} and c = Int32({{my_condition}}){% end %}", {'myvar': 1.0})
    'SELECT count() c FROM test_table where a > 1.0 '
    >>> render_sql_template("SELECT count() c FROM where {{sql_and(a=a, b=b)}}", {'a': '1', 'b': '2'})
    "SELECT count() c FROM where a = '1' and b = '2'"
    >>> render_sql_template("SELECT count() c FROM where {{sql_and(a=a, b=b)}}", {'b': '2'})
    "SELECT count() c FROM where b = '2'"
    >>> render_sql_template("SELECT count() c FROM where {{sql_and(a=Int(a, defined=False), b=Int(b, defined=False))}}", {'b': '2'})
    'SELECT count() c FROM where b = 2'
    >>> render_sql_template("SELECT count() c FROM where {{sql_and(a__in=Array(a), b=b)}}", {'a': 'a,b,c','b': '2'})
    "SELECT count() c FROM where a in ('a','b','c') and b = '2'"
    >>> render_sql_template("SELECT count() c FROM where {{sql_and(a__not_in=Array(a), b=b)}}", {'a': 'a,b,c','b': '2'})
    "SELECT count() c FROM where a not in ('a','b','c') and b = '2'"
    >>> render_sql_template("SELECT c FROM where a > {{Date(start)}}", test_mode=True)
    "SELECT c FROM where a > '2019-01-01'"
    >>> render_sql_template("SELECT c FROM where a > {{DateTime(start)}}", test_mode=True)
    "SELECT c FROM where a > '2019-01-01 00:00:00'"
    >>> render_sql_template("SELECT c FROM where a > {{DateTime(start)}}", {'start': '2018-09-07 23:55:00'})
    "SELECT c FROM where a > '2018-09-07 23:55:00'"
    >>> render_sql_template('SELECT * FROM tracker {% if defined(start) %} {{DateTime(start)}} and {{DateTime(end)}} {% end %}', {'start': '2019-08-01 00:00:00', 'end': '2019-08-02 00:00:00'})
    "SELECT * FROM tracker  '2019-08-01 00:00:00' and '2019-08-02 00:00:00' "
    >>> render_sql_template('SELECT * from test limit {{Int(limit)}}', test_mode=True)
    'SELECT * from test limit 0'
    >>> render_sql_template('SELECT {{symbol(attr)}} from test', test_mode=True)
    'SELECT `placeholder` from test'
    >>> render_sql_template('SELECT {{Array(foo)}}', {'foo': 'a,b,c,d'})
    "SELECT ('a','b','c','d')"
    >>> render_sql_template("SELECT {{Array(foo, 'Int32')}}", {'foo': '1,2,3,4'})
    'SELECT (1,2,3,4)'
    >>> render_sql_template("SELECT {{Array(foo, 'Int32')}}", test_mode=True)
    'SELECT (0,0)'
    >>> render_sql_template("SELECT {{Array(foo)}}", test_mode=True)
    "SELECT ('__placeholder__0','__placeholder__1')"
    >>> res = {}
    >>> render_sql_template("{{max_threads(2)}} SELECT 1", template_execution_results=res)
    '-- max_threads 2\\n SELECT 1'
    >>> res
    {'max_threads': 2}
    >>> render_sql_template("SELECT {{String(foo)}}", test_mode=True)
    "SELECT '__placeholder__'"
    >>> render_sql_template("SELECT {{String(foo, 'test')}}", test_mode=True)
    "SELECT 'test'"
    >>> render_sql_template("SELECT {{String(foo, 'test')}}", {'foo': 'tt'})
    "SELECT 'tt'"
    >>> render_sql_template("SELECT {{String(format, 'test')}}", {'format': 'tt'})
    Traceback (most recent call last):
    ...
    ValueError: "format" can not be used as a variable name, line 1
    >>> render_sql_template("SELECT {{format}}", {'format': 'tt'})
    Traceback (most recent call last):
    ...
    ValueError: "format" can not be used as a variable name, line 1
    >>> render_sql_template("SELECT {{String(q, 'test')}}", {'q': 'tt'})
    Traceback (most recent call last):
    ...
    ValueError: "q" can not be used as a variable name, line 1
    >>> render_sql_template("SELECT {{column(agg)}}", {})
    Traceback (most recent call last):
    ...
    ValueError: Missing column() default value, use `column(column_name, 'default_column_name')`
    >>> render_sql_template("SELECT {{column(agg)}}", {'agg': 'foo'})
    'SELECT `foo`'
    >>> render_sql_template('{% if not defined(test) %}error("This is an error"){% end %}', {})
    'error("This is an error")'
    >>> render_sql_template('{% if not defined(test) %}custom_error({error: "This is an error"}){% end %}', {})
    'custom_error({error: "This is an error"})'
    >>> render_sql_template("SELECT {{String(foo + 'abcd')}}", test_mode=True)
    "SELECT '__placeholder__'"
    >>> render_sql_template("SELECT {{columns(agg)}}", {})
    Traceback (most recent call last):
    ...
    ValueError: Missing columns() default value, use `columns(column_names, 'default_column_name')`
    >>> render_sql_template("SELECT {{columns(agg, 'a,b,c')}} FROM table", {})
    'SELECT `a`,`b`,`c` FROM table'
    >>> render_sql_template("SELECT {{columns(agg, 'a,b,c')}} FROM table", {'agg': 'foo'})
    'SELECT `foo` FROM table'
    >>> render_sql_template("SELECT {{columns('a,b,c')}} FROM table", {})
    'SELECT `a`,`b`,`c` FROM table'
    >>> render_sql_template("% {% if whatever(passenger_count) %}{% end %}", test_mode=True)
    Traceback (most recent call last):
    ...
    ValueError: 'whatever' is not a valid function, line 1
    >>> render_sql_template("% {% if defined((passenger_count) %}{% end %}", test_mode=True)
    Traceback (most recent call last):
    ...
    SyntaxError: invalid syntax
    >>> render_sql_template("SELECT * FROM dim_fecha_evento where foo like {{sql_unescape(String(pepe), '%')}}", {"pepe": 'raul_el_bueno_es_un_tikismikis_%'})
    "SELECT * FROM dim_fecha_evento where foo like 'raul_el_bueno_es_un_tikismikis_%'"
    """

    t = Template(sql)

    template_variables = get_var_names(t)

    for variable in template_variables:
        if variable['name'] in DEFAULT_PARAM_NAMES:
            name = variable['name']
            line = variable['line']
            raise ValueError(f'"{name}" can not be used as a variable name, line {line}')

    if test_mode:
        def dummy(*args, **kwargs):
            return Comment('error launched')
        v = {x['name']: Placeholder(x['name'], x['line']) for x in template_variables}

        if variables:
            v.update(variables)

        v.update(type_fns_check)
        v.update({
            # disable error throws on check
            'error': dummy,
            'custom_error': dummy
        })

    else:
        v = {x['name']: None for x in template_variables}
        if variables:
            v.update(variables)
        v.update(type_fns)

    wrap_vars(t)

    return generate(t, template_execution_results, **v).decode()



from pathlib import Path
import os.path
import json
import click

from tinybird import __cli__
from .feedback_manager import FeedbackManager

try:
    from tinybird.__cli__ import __revision__
except Exception:
    __revision__ = None

DEFAULT_API_HOST = 'https://api.tinybird.co'
DEFAULT_LOCALHOST = 'http://localhost:8001'
CURRENT_VERSION = f"{__cli__.__version__}"
VERSION = f"{__cli__.__version__} (rev {__revision__})"
DEFAULT_UI_HOST = 'https://ui.tinybird.co'
SUPPORTED_CONNECTORS = ['bigquery', 'snowflake']
PROJECT_PATHS = [
    'datasources',
    'datasources/fixtures',
    'endpoints',
    'explorations',
    'pipes'
]
MIN_WORKSPACE_ID_LENGTH = 36


async def get_config(hostFlag, tokenFlag):
    if hostFlag:
        hostFlag = hostFlag.rstrip('/')

    config_file = Path(os.getcwd()) / ".tinyb"
    config = {}
    try:
        with open(config_file) as file:
            config = json.loads(file.read())
    except IOError:
        pass
    except json.decoder.JSONDecodeError:
        click.echo(FeedbackManager.error_load_file_config(config_file=config_file))
        return
    config['token'] = tokenFlag or config.get('token', None)
    config['host'] = hostFlag or config.get('host', DEFAULT_API_HOST)
    config['workspaces'] = config.get('workspaces', [])
    return config


async def write_config(config, dest_file='.tinyb'):
    config_file = Path(os.getcwd()) / dest_file
    with open(config_file, 'w') as file:
        file.write(json.dumps(config, indent=4, sort_keys=True))


class FeatureFlags:
    @classmethod
    def ignore_sql_errors(cls) -> bool:  # Context: #1155
        return "TB_IGNORE_SQL_ERRORS" in os.environ

    @classmethod
    def is_localhost(cls) -> bool:
        return "SET_LOCALHOST" in os.environ




import logging
import re
import ast

datetime_patterns = [
    r'\d\d\d\d.\d\d.\d\d(T|\s)\d\d:\d\d:\d\d',
    r'\d\d.\d\d.\d\d\d\d.\d{1,2}:\d{1,2}:\d{1,2}'
]

int_8_max = 2**7
int16_max = 2**15
int32_max = 2**31
int64_max = 2**63
uint_8_max = 2**8
uint16_max = 2**16
uint32_max = 2**32
uint64_max = 2**64
intx_re = r'^[+-]?\d+$'
uintx_re = r'^\d+$'
float32_max = 2**23  # 23 bits is the fractional part of float 32 ieee754
float64_max = 2**52  # 51 bits is the fractional part of float 64 ieee754


def string_test(x):
    return True


def date_test(x):
    return re.match(r'\d\d\d\d-\d\d-\d\d$', x)


def datetime_test(x):
    return any([re.match(p, x) for p in datetime_patterns])


def int_8_test(x):
    return re.match(intx_re, x) and - int_8_max <= int(x) < int_8_max


def int16_test(x):
    return re.match(intx_re, x) and - int16_max <= int(x) < int16_max


def int32_test(x):
    return re.match(intx_re, x) and - int32_max <= int(x) < int32_max


def int64_test(x):
    return re.match(intx_re, x) and - int64_max <= int(x) < int64_max


def uint_8_test(x):
    return re.match(uintx_re, x) and 0 <= int(x) < uint_8_max


def uint16_test(x):
    return re.match(uintx_re, x) and 0 <= int(x) < uint16_max


def uint32_test(x):
    return re.match(uintx_re, x) and 0 <= int(x) < uint32_max


def uint64_test(x):
    return re.match(uintx_re, x) and 0 <= int(x) < uint64_max


def float_test(x):
    return '_' not in x and type_test(x, float)


def float32_test(x):
    return '_' not in x and type_test(x, float) and -float32_max <= float(x) < float32_max


def float64_test(x):
    return '_' not in x and type_test(x, float) and -float64_max < float(x) < float64_max


def test_numeric_testers(fn, n):
    """
    >>> test_numeric_testers(int32_test, (2**31)-1)
    True
    >>> test_numeric_testers(int32_test, -(2**31))
    True
    >>> test_numeric_testers(int32_test, -(2**31)-1)
    False
    >>> test_numeric_testers(int32_test, 2**31)
    False

    >>> test_numeric_testers(int64_test, (2**63)-1)
    True
    >>> test_numeric_testers(int64_test, -(2**63))
    True
    >>> test_numeric_testers(int64_test, -(2**63)-1)
    False
    >>> test_numeric_testers(int64_test, 2**63)
    False
    """
    return fn(str(n))


def array_test(_type_test):
    """
    >>> array_test(str)("['blabla']")
    True
    >>> array_test(str)('["blabla"]')
    True
    >>> array_test(str)('["blabla","bloblo"]')
    True
    >>> array_test(str)('["blabla, bloblo"]')
    True
    >>> array_test(str)("[ W ]")
    False
    >>> array_test(int)("[1]")
    True
    >>> array_test(int)('[1]')
    True
    >>> array_test(int)('[1,2]')
    True
    >>> array_test(float)("[1.2]")
    True
    >>> array_test(float)('[1.2]')
    True
    >>> array_test(float)('[1.2,2.1]')
    True
    >>> array_test(float)('["1.2","2.1"]')
    False
    """
    def _test(x):
        if x[0] != '[':
            return False
        try:
            k = ast.literal_eval(x)
        except Exception as error:
            logging.warning('array test warning: ' + str(error))
            return False
        if isinstance(k, list):
            return all(isinstance(x, _type_test) for x in k)
        return False
    return _test


numbers_types = (
    'Int8',
    'UInt8',
    'Int16',
    'UInt16',
    'UInt32',
    'Int32',
    'Int64',
    'UInt64',
    'Float32',
    'Float64'
)

# Use guessers for discovering types
# I.e., when you have to take into consideration things like float precision
guessers = {
    'DateTime': datetime_test,
    'Date': date_test,
    'Int8': int_8_test,
    'UInt8': uint_8_test,
    'Int16': int16_test,
    'UInt16': uint16_test,
    'Int32': int32_test,
    'UInt32': uint32_test,
    'Int64': int64_test,
    'UInt64': uint64_test,
    'Float32': float32_test,
    'Float64': float64_test,
    'Array(Int32)': array_test(int),
    'Array(Float32)': array_test(float),
    'Array(String)': array_test(str)
}

# Use testers validating a value against a type
# I.e., you already know the type and you need to check if a value fits there
testers = {
    'DateTime': datetime_test,
    'Date': date_test,
    'Int8': int_8_test,
    'UInt8': uint_8_test,
    'Int16': int16_test,
    'UInt16': uint16_test,
    'Int32': int32_test,
    'UInt32': uint32_test,
    'Int64': int64_test,
    'UInt64': uint64_test,
    'Float32': float_test,
    'Float64': float_test,
    'Array(Int32)': array_test(int),
    'Array(Float32)': array_test(float),
    'Array(String)': array_test(str)
}


def type_test(i, t):
    try:
        t(i)
        return True
    except Exception:
        return False
